/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * dmraid device-mapper lib interface functions.
 */

#include <libdevmapper.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>

#include "internal.h"
#include "devmapper.h"

/* Wrap dm_dir() */
const char *get_dm_dir(void)
{
	return dm_dir();
}

/* Make a dm path. */
char *mkdm_path(struct lib_context *lc, const char *name)
{
	char *ret;
	const char *dir = get_dm_dir();

	if (!(ret = dbg_malloc(strlen(dir) + strlen(name) + 2)))
		LOG_ERR(lc, NULL, "Failed allocating dm name");

	sprintf(ret, "%s/%s", dir, name);
	return ret;
}

/*
 * Parse a mapping table and create the appropriate targets.
 */
static int parse_table(struct lib_context *lc, struct dm_task *dmt, char *table)
{
	int ret = 0, n, line = 0;
	unsigned long long start, size;
	char *nl, *p = table, ttype[32];

	while (*p) {
		line++;

		/*
		 * Not using sscanf string allocation
		 * because of its lag in dietlibc.
		 */
		memset(ttype, 0, 32);
		if (sscanf(p, "%llu %llu %31s %n",
			   &start, &size, ttype, &n) < 3)
			LOG_ERR(lc, 0, "Invalid format in table line %d", line);

		p += n;
		if ((nl = strchr(p, '\n')))
			*nl = '\0';

		ret = dm_task_add_target(dmt, start, size, ttype, p);
		if (!ret)
			goto out;

		p += strlen(p);
		if (nl) {
			*nl = '\n';
			p++;
		}
	}

	ret = 1;

  out:
	return ret;
}

/* Device-mapper NULL log function. */
static void dmraid_log(int level, const char *file, int line,
		       const char *f, ...)
{
	return;
}

/* Init device-mapper library. */
static void _init_dm(void)
{
	dm_log_init(dmraid_log);
}

/* Cleanup at exit. */
static void _exit_dm(struct dm_task *dmt)
{
	if (dmt)
		dm_task_destroy(dmt);

	dm_lib_release();
	dm_lib_exit();
}

/* Create a mapped device. */
int dm_create(struct lib_context *lc, struct dmraid_set *rs, char *table)
{
	int ret = 0;
	struct dm_task *dmt;

	_init_dm();

	/* Create <dev_name> */
	if (!(dmt = dm_task_create(DM_DEVICE_CREATE)))
		goto out;

	if (!dm_task_set_name(dmt, rs->name))
		goto out;

	if (!parse_table(lc, dmt, table))
		goto out;

	if (!dm_task_run(dmt))
		goto out;

	ret = 1;

   out:
	_exit_dm(dmt);
	return ret;
}

/* Remove a mapped device. */
int dm_remove(struct lib_context *lc, struct dmraid_set *rs)
{
	int ret = 0;
	struct dm_task *dmt;

	_init_dm();

	/* remove <dev_name> */
	if (!(dmt = dm_task_create(DM_DEVICE_REMOVE)))
		goto out;

	if (!dm_task_set_name(dmt, rs->name))
		goto out;

	ret = dm_task_run(dmt);

   out:
	_exit_dm(dmt);
	return ret;
}

/* Retrieve status of a mapped device. */
/* FIXME: more status... */
int dm_status(struct lib_context *lc, struct dmraid_set *rs)
{
	int ret = 0;
	struct dm_task *dmt;
	struct dm_info info;

	_init_dm();

	/* Status <dev_name> */
	if (!(dmt = dm_task_create(DM_DEVICE_STATUS)))
		goto out;

	if (!dm_task_set_name(dmt, rs->name))
		goto out;

	if (!dm_task_run(dmt))
		goto out;

	if (!dm_task_get_info(dmt, &info))
		goto out;

	if (info.exists)
		ret = 1;

   out:
	_exit_dm(dmt);
	return ret;
}

/* Retrieve device-mapper driver version. */
int dm_version(struct lib_context *lc, char *version, size_t size)
{
	int ret = 0;
	struct dm_task *dmt;

	strncpy(version, "unknown", size);
	_init_dm();

	if (!(dmt = dm_task_create(DM_DEVICE_VERSION)))
		goto out;

	if (!dm_task_run(dmt))
		goto out;

	if (!dm_task_get_driver_version(dmt, version, size))
		goto out;

	ret = 1;

  out:
	_exit_dm(dmt);
	return ret;
}
