/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include <dirent.h>
#include <mntent.h>
#include <stdlib.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include "internal.h"
#include "scsi.h"

/*
 * subdirectory below sysfs moint point holding the 
 * subdirectory hierarchies of all block devices.
 */
#define	BLOCK 		"/block"

/* Find sysfs mount point */
#ifndef	_PATH_MOUNTS
#define	_PATH_MOUNTS	"/proc/mounts"
#endif
static char *find_sysfs_mp(struct lib_context *lc)
{
	char *ret = NULL;
	FILE *mfile;
	struct mntent *ment;

	/* Try /proc/mounts and fallback to /etc/mtab. */
	if (!(mfile = setmntent(_PATH_MOUNTS, "r"))) {
		if (!(mfile = setmntent(_PATH_MOUNTED, "r")))
			LOG_ERR(lc, NULL, "Unable to open %s or %s",
				_PATH_MOUNTS, _PATH_MOUNTED);
	}

	while ((ment = getmntent(mfile))) {
		if (!strcmp(ment->mnt_type, "sysfs")) {
			ret = ment->mnt_dir;
			break;
		}
	};

	endmntent(mfile);
	return ret;
}

/* Make up an absolute sysfs path given a relative one */
static char *mk_sysfs_path(struct lib_context *lc, char const *path) {
	char *ret = NULL;
	char *sysfs_mp;

	if (!(sysfs_mp = find_sysfs_mp(lc)))
		LOG_ERR(lc, NULL, "Failed to find sysfs mount point");

	if (!(ret = dbg_malloc(strlen(sysfs_mp) + strlen(path) + 1)))
		LOG_ERR(lc, NULL, "Failed to allocate sysfs path");

	sprintf(ret, "%s%s", sysfs_mp, path);
	return ret;
}

/* Allocate disk_info struct and keep the device path */
struct disk_info *alloc_disk_info(char *path)
{
	struct disk_info *di;

	if ((di = dbg_malloc(sizeof(*di)))) {
		memset(di, 0, sizeof(*di));
		if ((di->path = dbg_strdup(path)))
			INIT_LIST_HEAD(&di->list);
		else {
			dbg_free(di);
			di = NULL;
		}
	}

	return di;
}

/* Free disk_info structure */
static void _free_disk_info(struct lib_context *lc, struct disk_info *di)
{
	if (di->rd)
		free_dev(lc, di->rd);

	if (di->serial)
		dbg_free(di->serial);

	dbg_free(di->path);
	dbg_free(di);
}

/* Free disk_info structure or all registered disk_infos in case di = NULL. */
void free_disk_info(struct lib_context *lc, struct disk_info *di)
{
	if (di)
		_free_disk_info(lc, di);
	else {
		struct list_head *elem, *tmp;
	
		list_for_each_safe(elem, tmp, lc_list(lc, LC_DISK_INFOS)) {
			list_del(elem);
			_free_disk_info(lc,
					list_entry(elem, struct disk_info,
					list));
		}
	
		if (!list_empty(lc_list(lc, LC_DISK_INFOS)))
			log_fatal(lc, "disk_info list not empty");
	}
}

/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
static int dm_device(char *name)
{
	return !strncmp(name, "dm-", 3);
}

/* Fake a SCSI serial number by reading it from a file. */
static int get_dm_serial(struct lib_context *lc,
			 struct disk_info **di, char *path)
{
	int i, ret = 0;
	FILE *f;
	char *serial, *str, buffer[64];

	if (!(serial = dbg_malloc(strlen(path) + strlen(".serial") + 1)))
		LOG_ERR(lc, 0, "Failed allocating dm device serial path for %s",
			path);

	sprintf(serial, "%s.serial", path);
	if (!(f = fopen(serial, "r")))
		goto free;

	if (!fgets(buffer, 32, f))
		goto out;

	for (i = strlen(buffer) - 1; i > 0 && isspace(buffer[i]); i--)
		buffer[i] = '\0';
	str = buffer;
	while (isspace(*str))
		str++;

	if (((*di)->serial = dbg_strdup(str)))
		ret = 1;

  out:
	fclose(f);
  free:
	dbg_free(serial);

	return ret;
}
#endif

/*
 * Create device node path, allocate disk_info and ioctl for sector size.
 */
static int alloc_di_ioctl(struct lib_context *lc,
			  struct disk_info **di, char *name)
{
	int fd, ret = 0, sector_size = 0;
	char *path;

	if (!(path = dbg_malloc(strlen(name) + strlen(_PATH_DEV) + 1)))
		LOG_ERR(lc, 0, "Failed allocating block device path");

	sprintf(path, "%s%s", _PATH_DEV, name);
	if ((fd = open(path, O_RDONLY)) == -1)
		goto free;

	if (!(*di = alloc_disk_info(path)))
		goto close;

	if (ioctl(fd, BLKSSZGET, &sector_size))
		sector_size = DMRAID_SECTOR_SIZE;

	if (sector_size != DMRAID_SECTOR_SIZE) {
		log_err(lc, "unsupported sector size %d on %s.", sector_size, path);
		goto free1;
	}

	/* REMOVEME: Test with dm devices ? */
#ifdef	DMRAID_TEST
	if (dm_device(name)) {
		if (!get_dm_serial(lc, di, path))
			log_warn(lc, "missing dm serial file for %s", path);
	} else {
#endif
		get_scsi_serial(di, fd);
#ifdef	DMRAID_TEST
	}
#endif

	ret = 1;
	goto close;

   free1:
	_free_disk_info(lc, *di);
	*di = NULL;
   close:
	close(fd);
    free:
	dbg_free(path);

	return ret;
}

/* Read the size in sectors from the sysfs "size" file */
static int _get_size(struct lib_context *lc,
		     struct disk_info **di, const char *path, char *name)
{
	int ret = 0;
	char *sysfs_file;
	FILE *f;

	if (!(sysfs_file = dbg_malloc(strlen(path) + strlen(name) + 7)))
		LOG_ERR(lc, 0, "Failed allocating sysfs block device path");

	sprintf(sysfs_file, "%s/%s/size", path, name);
	if (!(f = fopen(sysfs_file, "r"))) {
		log_err(lc, "failed opening %s", sysfs_file);
		goto free;
	}

	if (fscanf(f, "%" PRIu64, &(*di)->sectors) != 1)
		log_err(lc, "failed reading disk size for %s from sysfs", path);
	else
		ret = 1;

	fclose(f);

   free:
	dbg_free(sysfs_file);
	return ret;
}

/* Support function to adjust crappy disk geometries */
/* Are we interested in this device ? */
static int interested(char *name)
{
	/*
	 * Whole IDE and SCSI disks only.
	 * Include dm devices for testing.
	 */
	return ((!isdigit(name[strlen(name)-1]) &&
	        *(name+1) == 'd' &&
	        (*name == 'h' || *name == 's'))
#ifdef	DMRAID_TEST
		|| dm_device(name)
#endif
	       ) ? 1 : 0;
}

static int get_size(struct lib_context *lc, const char *path,
		    char *name, int sysfs)
{
	struct disk_info *di = NULL;

	if (!interested(name))
		return 1;

	if (!alloc_di_ioctl(lc, &di, name) ||
	    (sysfs && !_get_size(lc, &di, path, name))) {
		if (di) {
			_free_disk_info(lc, di);
			return 1;
		}

		return 0;
	}

	list_add(&di->list, lc_list(lc, LC_DISK_INFOS));
	return 1;
}

/*
 * Find disk devices in sysfs or directly
 * in /dev on Linux 2.4) and keep information.
 */
int discover_devices(struct lib_context *lc)
{
	int sysfs = 1, ret = 0;
	char *path, *p;
	DIR *d;
	struct dirent *de;

	if ((p = mk_sysfs_path(lc, BLOCK)))
		path = p;
	else {
		path = (char*) _PATH_DEV;
		sysfs = 0;
		log_print(lc, "carrying on with %s", path);
	}

	if (!(d = opendir(path))) {
		log_err(lc, "failed opening path %s", path);
		goto out;
	}

	while ((de = readdir(d)))
		get_size(lc, path, de->d_name, sysfs);

	closedir(d);
	ret = 1;

  out:
	if (p)
		dbg_free(p);

	return ret;
}
