/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include "internal.h"

/* Display information about a block device */
static void log_disk(struct lib_context *lc, struct disk_info *di)
{
	const char *fmt;

	switch (lc_opt(lc, LC_COLUMN)) {
	case 0:
		fmt = "%s: %12" PRIu64 " total, \"%s\"";
		break;

	case 1:
		fmt = "%s";
		break;

	default:
		fmt = "%s,%" PRIu64 ",\"%s\"";
	}

	log_print(lc, fmt,
		  di->path, di->sectors, di->serial ? di->serial : "N/A");
}

/* Display information about a RAID device */
static void log_rd(struct lib_context *lc, struct dmraid_dev *rd)
{
	const char *fmt;

	if (!rd)
		return;

	if (rd->sectors) {
		switch (lc_opt(lc, LC_COLUMN)) {
			case 0:
				fmt = "%s: %-18s, \"%s\", %7s, %" PRIu64
				      " sectors, data@ %" PRIu64;
				break;
	
			case 1:
				fmt = "%s";
				break;
	
			default:
				fmt = "%s,%s,%s,%s,%" PRIu64 ",%" PRIu64;
		}
	} else {
		switch (lc_opt(lc, LC_COLUMN)) {
			case 0:
				fmt = "%s: \"%s\" format, not used";
				break;
	
			default:
				fmt = "%s,%s,free";
		}
	}

	if (rd->sectors) {
		const char *type =
			rd->fmt->type ? get_type(lc, rd->fmt->type(lc, rd)) :
					"unsupported";

		if (!type)
			type = "unknown";

		log_print(lc, fmt, rd->di->path, rd->name, rd->fmt->name,
			  type, rd->sectors, rd->offset);
	} else
		log_print(lc, fmt, rd->di->path, rd->fmt->name);
}

/* Log native RAID device information. */
static void log_rd_native(struct lib_context *lc, struct dmraid_dev *rd)
{
	if (!rd)
		return;

	if (rd->fmt->log) {
		rd->fmt->log(lc, rd);
		log_print(lc, " ");
	} else
		log_print(lc, "\"%s\" doesn't support native logging of RAID "
			  "device information", rd->fmt->name);
}

/* Dispatch log functions. */
static void log_device(struct lib_context *lc, struct disk_info *di,
		       enum dev_type type)
{
	if (type == RAID)
		log_rd(lc, di->rd);
	else if (type == NATIVE)
		log_rd_native(lc, di->rd);
	else
		log_disk(lc, di);
}

/* Display information about a dmraid format handler */
static void log_format(struct lib_context *lc, struct dmraid_format *fmt)
{
	log_print(lc, "%-6s : %s", fmt->name, fmt->descr);
}

/* Display information about found block devices */
static void process_dis(void (*f)(struct lib_context *lc, struct disk_info *di,
				  enum dev_type arg),
				  struct lib_context *lc, enum dev_type type)
{
	struct disk_info *di;

	list_for_each_entry(di, lc_list(lc, LC_DISK_INFOS), list)
		f(lc, di, type);
}

void display_devices(struct lib_context *lc, enum dev_type type)
{
	int devs;

	if ((devs = count_devices(lc, type))) {
		log_notice(lc, "%s device%s discovered:\n",
			   (type == RAID || type == NATIVE) ? "RAID" : "Block",
			   devs == 1 ? "" : "s");
		process_dis(log_device, lc, type);
	}
}

static void _display_set(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int devs = 0;
	uint64_t sectors = 0;
	struct dmraid_dev *rd;
	const char *fmt, *type, *status;

	if (rs->type == t_group && !lc_opt(lc, LC_GROUP))
		return;

	if (!(type = get_set_type(lc, rs)))
		type = "unknown";

	if (!(status = get_status(lc, rs)))
		status = "unknown";

	switch (lc_opt(lc, LC_COLUMN)) {
	case 0:
		fmt = "name  : %s\n"
		      "size  : %" PRIu64 "\n"
		      "stride: %u\n"
		      "type  : %s\n"
		      "status: %s\n"
		      "dev   : %d";
		devs = count_devs(lc, rs);
		sectors = calc_total_sectors(lc, rs);
		break;

	case 1:
		fmt = "%s";
		break;

	default:
		fmt = "%s,%" PRIu64 ",%u,%s,%s,%u";
		devs = count_devs(lc, rs);
		sectors = calc_total_sectors(lc, rs);
	}

	log_print(lc, fmt, rs->name, sectors, rs->stride, type, status, devs);

	if (lc_opt(lc, LC_COLUMN) > 2) {
		list_for_each_entry(rd, &rs->devs, list)
			log_rd(lc, rd);
	}
}

/* FIXME: Spock, do something better (neater). */
static int group_active(struct lib_context *lc, struct dmraid_set *rs)
{
	struct dmraid_set *r;

	list_for_each_entry(r, &rs->sets, list) {
		if (dm_status(lc, r))
			return 1;
	}

	return 0;
}

void display_set(struct lib_context *lc, void *v,
		 enum active_type active, int top)
{
	struct dmraid_set *rs = v;
	struct dmraid_set *r;
	int dmstatus = rs->type == t_group ?
		       group_active(lc, rs) : dm_status(lc, rs);

	if ((active == A && !dmstatus) ||
	   (active == I && dmstatus))
		return;

	if (!lc_opt(lc, LC_COLUMN)) {
		if (rs->type == t_group && !lc_opt(lc, LC_GROUP))
			log_print(lc, "*** Group superset %s", rs->name);
		else {
			log_print(lc, "%s %s%s%set",
		  		top ? "--->" : "\n***",
		  		rs->status == s_inconsistent ?
					      "*Inconsistent* " : "",
		  		dm_status(lc, rs) ? "Active " : "",
		  		!list_empty(&rs->sets) ?
				"Supers" : (top ? "Subs" : "S"));
		}
	}

	_display_set(lc, rs);

	list_for_each_entry(r, &rs->sets, list)
		display_set(lc, r, active, ++top);
}

/*
 * Display information about supported RAID metadata formats
 * (ie. registered format handlers)
 */
static void _list_formats(struct lib_context *lc, enum fmt_type type)
{
	struct format_list *fmt_list;

	list_for_each_entry(fmt_list, lc_list(lc, LC_FORMATS), list) {
		if (type == fmt_list->fmt->format)
			log_format(lc, fmt_list->fmt);
	}
}

int list_formats(struct lib_context *lc, int arg)
{
	if (!list_empty(lc_list(lc, LC_FORMATS)))
		log_notice(lc, "supported Metadata Formats:\n");

	_list_formats(lc, FMT_RAID);
	_list_formats(lc, FMT_PARTITION);

	log_print(lc, "(0): Discover, (+): Discover+Activate");
	return 1;
}
