/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Highpoint 37X ATARAID metadata format handler.
 *
 * hpt37x_read(), hpt37x_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "hpt37x.h"
#include <datastruct/byteorder.h>

static const char *handler = "hpt37x";

/* Make up RAID set name from magic_[01] numbers */
/* FIXME: better name ? */
static size_t __name(struct hpt37x *hpt, char *str, size_t len,
		     unsigned char type)
{
	const char *fmt = (hpt->magic_1 && type) ? "hpt37x_%u-%u" : "hpt37x_%u";
		
	return snprintf(str, len, fmt,
			hpt->magic_0, hpt->magic_1) + 1;
}

static char *_name(struct lib_context *lc, struct hpt37x *hpt,
		   unsigned char type)
{
	size_t len;
	char *ret;

	len  = __name(hpt, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(hpt, ret, len, type);
		mk_alpha(ret + 7, strlen(ret) - 7);
	} else
		log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status _status(struct dmraid_dev *rd)
{
	struct hpt37x* hpt = (struct hpt37x*) rd->meta_areas[0].area;

	return hpt->magic == HPT37X_MAGIC_BAD ? s_broken : s_operational;
}

/* Mapping of HPT 37X types to generic types. */
static struct types types[] = {
	{ HPT37X_T_SINGLEDISK, t_linear},
	{ HPT37X_T_SPAN, t_linear},
	{ HPT37X_T_RAID0, t_raid0},
	{ HPT37X_T_RAID1, t_raid1},
	{ HPT37X_T_RAID01_RAID0, t_raid1},
	/* FIXME: support RAID 3+5 */
	{ 0, t_undef}
};

/* Neutralize disk type. */
static enum type hpt37x_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type(types, (unsigned int) ((struct hpt37x*) rd->meta_areas[0].area)->type);
}

/*
 * Decide about ordering sequence of RAID device.
 * Called by list_add_dev_sorted().
 */
static int _hpt37x_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct hpt37x*) new->meta_areas[0].area)->disk_number <
	       ((struct hpt37x*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Read a Highpoint 37X RAID device.
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct hpt37x *hpt)
{
	struct hpt37x_errorlog *l;

	hpt->magic	= le32_to_cpu(hpt->magic);
	hpt->magic_0	= le32_to_cpu(hpt->magic_0);
	hpt->magic_1	= le32_to_cpu(hpt->magic_1);
	hpt->order	= le32_to_cpu(hpt->order);
	hpt->total_secs	= le32_to_cpu(hpt->total_secs);
	hpt->disk_mode	= le32_to_cpu(hpt->disk_mode);
	hpt->boot_mode	= le32_to_cpu(hpt->boot_mode);

	for (l = &hpt->errorlog[0];
	     l < &hpt->errorlog[hpt->error_log_entries];
	     l++) {
		l->timestamp = le32_to_cpu(l->timestamp);
		l->lba = le32_to_cpu(l->lba);
	}
}
#endif

/* Magic check. */
static int hpt37x(struct hpt37x *hpt)
{
	return (hpt->magic == HPT37X_MAGIC_OK ||
		hpt->magic == HPT37X_MAGIC_BAD) &&
		hpt->disk_number < 8;
}

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct hpt37x *hpt, struct disk_info *di);
static struct dmraid_dev *hpt37x_read(struct lib_context *lc,
				      struct disk_info *di)
{
	struct hpt37x *hpt;
	struct dmraid_dev *rd = NULL;

	if (!(hpt = alloc_private(lc, handler, sizeof(*hpt))))
		goto out;

	if (!read_file(lc, di->path, (void *) hpt, sizeof(*hpt),
		       HPT37X_CONFIGOFFSET))
		goto bad;

	to_cpu(hpt);
	if (!hpt37x(hpt))
		goto bad;

	file_metadata(lc, handler, di->path, hpt, sizeof(*hpt),
		      HPT37X_CONFIGOFFSET);

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, hpt, di))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(hpt);
   out:
	return rd;
}

/*
 * Write a Highpoint 37X RAID device.
 */
static int hpt37x_write(struct lib_context *lc,
			struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct hpt37x *hpt = rd->meta_areas[0].area;
#endif

	to_disk(hpt);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(hpt);

	return ret;
}

/*
 * Group the RAID disk into a set.
 *
 * Check device hierarchy and create sub sets appropriately.
 *
 */
static unsigned int stride(struct hpt37x *hpt)
{
	return hpt->raid0_shift ? 1 << hpt->raid0_shift : 0;
}

static int _alloc_subsets(struct lib_context *lc, struct dmraid_set *rs,
			  struct dmraid_dev *rd, struct hpt37x *hpt)
{
	unsigned int i;
	struct dmraid_set *r;

	/* Allocate 2 subsets and link them to the superset */
	for (i = 0; i < 2; i++) {
		if ((r = alloc_set(lc))) {
			r->type = (hpt->type == HPT37X_T_RAID01_RAID0) ? t_raid0 : t_raid1;
			r->status = s_setup;
			list_add_tail(&r->list, &rs->sets);
		} else {
			if (i) {
				r = list_entry(rs->sets.next, typeof(*r), list);
				list_del(&r->list);
				/* Set is virgin so we can just free it here. */
				dbg_free(r);
				INIT_LIST_HEAD(&rs->sets);
			}

			return 0;
		}
	}

	/* Set properties of superset */
	rs->name = _name(lc, hpt, 0);
	rs->type = (hpt->type == HPT37X_T_RAID01_RAID0) ? t_raid1 : t_raid0;

	return 1;
}

static int group_rd(struct lib_context *lc, struct dmraid_set *rs,
		    struct dmraid_dev *rd)
{
	unsigned int sorted;
	struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area, *h = NULL;
	struct dmraid_set *r;

	if (rd_type(types, hpt->type) == t_undef)
		LOG_ERR(lc, 0, "%s: RAID type %u not supported",
			handler, hpt->type);

	if (!list_empty(&rs->devs))
		h = (struct hpt37x*) list_entry(rs->devs.next, typeof(*rd), list)->meta_areas[0].area;

	rs->stride = stride(hpt);

	switch (hpt->type) {
	case HPT37X_T_RAID01_RAID0:
	case HPT37X_T_RAID01_RAID1:
		if (h && h->magic_1 != hpt->magic_1)
			LOG_ERR(lc, 0, "%s: magic_1 mismatch", handler);

		/* Allocate 2 subsets for the stripe/mirror on top */
		if (list_empty(&rs->sets)) {
			if (!_alloc_subsets(lc, rs, rd, hpt))
				return 0;
		}

		/* Sort RAID disk into appropriate subset */
		sorted = 0;
		list_for_each_entry(r, &rs->sets, list) {
			if (list_empty(&r->devs)) {
				if (!(rs->name = dbg_strdup(rd->name)))
					return 0;
			} else
				h = (struct hpt37x*) list_entry(r->devs.next, typeof(*rd), list)->meta_areas[0].area;

			if (list_empty(&r->devs) ||
			    h->magic_0 == hpt->magic_0) {
				list_add_dev_sorted(lc, r, rd, _hpt37x_sort);
				sorted++;
				break;
			}
		}

		if (!sorted)
			goto out;

		break;

	default:
		if (h && h->magic_0 != hpt->magic_0)
			goto out;

		list_add_dev_sorted(lc, rs, rd, _hpt37x_sort);
		break;
	}

	return 1;

  out:
	log_err(lc, "%s: magic_0 mismatch", handler);
	return 0;
}

/*
 * Add a Highpoint RAID device to a set.
 */
static struct dmraid_set *hpt37x_group(struct lib_context *lc,
				        struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	if (!(rs = find_set(lc, rd->name, 0))) {
		if (!(rs = alloc_set(lc)))
			return NULL;

		if (!(rs->name = dbg_strdup(rd->name))) {
			dbg_free(rs);
			return NULL;
		}

		rs->type = hpt37x_type(lc, rd);
		rs->status = s_setup;
		list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));
	}

	return group_rd(lc, rs, rd) ? rs : NULL;
}

/*
 * Check a Highpoint 37X RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _hpt37x_check(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;

	if (list_empty(&rs->devs))
		return;

	devs = count_devs(lc, rs);
	list_for_each_entry(rd, &rs->devs, list) {
		struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area;

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		if (devs != hpt->raid_disks) {
			rd->status = s_broken;
			log_err(lc, "%s: device %s broken in RAID "
				"set \"%s\" [%u/%u]",
				handler, rd->di->path, rs->name,
				devs, hpt->raid_disks);
		}
	}
}

static int hpt37x_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Check any hierachical RAID sets. */
	check_rs(lc, rs, hpt37x_check);

	/* Check the lowest level RAID devices. */
	_hpt37x_check(lc, rs);

	return set_rs_status(lc, rs) == s_broken ? 0 : 1;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an HPT37X RAID device.
 */
static void hpt37x_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int i;
	struct hpt37x *hpt = (struct hpt37x*) rd->meta_areas[0].area;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("magic: 0x%x", hpt, hpt->magic);
	DP("magic_0: 0x%x", hpt, hpt->magic_0);
	DP("magic_1: 0x%x", hpt, hpt->magic_1);
	DP("order: %u", hpt, hpt->order);
	DP("raid_disks: %u", hpt, hpt->raid_disks);
	DP("raid0_shift: %u", hpt, hpt->raid0_shift);
	DP("type: %u", hpt, hpt->type);
	DP("disk_number: %u", hpt, hpt->disk_number);
	DP("total_secs: %u", hpt, hpt->total_secs);
	DP("disk_mode: 0x%x", hpt, hpt->disk_mode);
	DP("boot_mode: 0x%x", hpt, hpt->boot_mode);
	DP("boot_disk: %u", hpt, hpt->boot_disk);
	DP("boot_protect: %u", hpt, hpt->boot_protect);
	DP("error_log_entries: %u", hpt, hpt->error_log_entries);
	DP("error_log_index: %u", hpt, hpt->error_log_index);
	if (hpt->error_log_entries)
		log_print(lc, "error_log:");

	for (i = 0; i < 32; i++) {
		if (!hpt->errorlog[i].timestamp)
			break;

		DP("timestamp: %u", hpt, hpt->errorlog[i].timestamp);
		DP("reason: %u", hpt, hpt->errorlog[i].reason);
		DP("disk: %u", hpt, hpt->errorlog[i].disk);
		DP("status: %u", hpt, hpt->errorlog[i].status);
		DP("sectors: %u", hpt, hpt->errorlog[i].sectors);
		DP("lba: %u", hpt, hpt->errorlog[i].lba);
       };
}
#endif

static struct dmraid_format hpt37x_format = {
	.name	= "hpt37x",
	.descr	= "(+) Highpoint HPT37X",
	.format = FMT_RAID,
	.read	= hpt37x_read,
	.write	= hpt37x_write,
	.group	= hpt37x_group,
	.check	= hpt37x_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= hpt37x_log,
#endif
	.type	= hpt37x_type,
};

/* Register this format handler with the format core */
int register_hpt37x(struct lib_context *lc)
{
	return register_format_handler(lc, &hpt37x_format);
}

/* Derive the RAID device contents from the Highpoint ones. */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct hpt37x *hpt, struct disk_info *di)
{
	rd->di = di;
        rd->name = _name(lc, hpt, 1);
	rd->fmt = &hpt37x_format;

	/* Data offset from start of device; first device is special */
	rd->offset = hpt->disk_number ? HPT37X_DATAOFFSET : 0;
        rd->sectors = hpt->total_secs;
        rd->sectors += rd->sectors & 1;
        rd->sectors /= hpt->raid_disks;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = HPT37X_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*hpt);
	rd->meta_areas[0].area = (void*) hpt;

	rd->status = _status(rd);

	return 1;
}
