/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Highpoint 45X ATARAID metadata format handler.
 *
 * hpt45x_read(), hpt45x_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "hpt45x.h"
#include <datastruct/byteorder.h>

static const char *handler = "hpt45x";

/* Make up RAID set name from magic_0 number */
/* FIXME: better name ? */
static size_t __name(struct hpt45x *hpt, char *str, size_t len,
		     unsigned char type)
{
	const char *fmt = type ? "hpt45x_%u-%u" : "hpt45x_%u";

	return snprintf(str, len, fmt, hpt->magic_0,
			hpt->raid1_disk_number) + 1;
}

static char *_name(struct lib_context *lc, struct hpt45x *hpt,
		   unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(hpt, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(hpt, ret, len, type);
		mk_alpha(ret + 7, strlen(ret) - 7);
	} else
		log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status _status(struct dmraid_dev *rd)
{
	struct hpt45x* hpt = (struct hpt45x*) rd->meta_areas[0].area;

	return hpt->magic == HPT45X_MAGIC_BAD ? s_broken : s_operational;
}

/* Mapping of HPT 45X types to generic types */
static struct types types[] = {
	{ HPT45X_T_SPAN, t_linear},
	{ HPT45X_T_RAID0, t_raid0},
	{ HPT45X_T_RAID1, t_raid1},
/* FIXME: handle RAID 3+5 */
	{ 0, t_undef}
};

/* Neutralize disk type */
static enum type hpt45x_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type(types,
		       (unsigned int) ((struct hpt45x*) rd->meta_areas[0].area)->type);
}

/*
 * Decide about ordering sequence of RAID device.
 * Called by list_add_dev_sorted().
 */
static int _hpt45x_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct hpt45x*) new->meta_areas[0].area)->disk_number <
	       ((struct hpt45x*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Group the RAID disk into a HPT45X set.
 *
 * Check device hierarchy and create super set appropriately.
 */
static unsigned int stride(unsigned int shift)
{
	return shift ? 1 << shift : 0;
}

static int group_rd(struct lib_context *lc, struct dmraid_set *rs,
		    struct dmraid_set **super_set, struct dmraid_dev *rd)
{
	unsigned int sorted;
	char *name;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;
	struct hpt45x *h = NULL;
	struct dmraid_set *r, *super;

	if (rd_type(types, hpt->type) == t_undef)
		LOG_ERR(lc, 0, "%s: RAID type %u not supported",
			handler, hpt->type);

	if (!list_empty(&rs->devs))
		h = (struct hpt45x*) list_entry(rs->devs.next, typeof(*rd), list)->meta_areas[0].area;

	switch (hpt->type) {
	case HPT45X_T_RAID0:
		if (hpt->raid1_type != HPT45X_T_RAID1)
			goto no_raid10;
		/*
		 * We've got a striped raid set with a mirror on top
		 * when we get here.
		 * Let's find and optionally allocate the mirror set on top.
		 */

		/* Superset name */
		if (!(name = _name(lc, hpt, 0)))
			return 0;

		if ((super = find_set(lc, name, 0)))
			dbg_free(name);
		else {
			if (!(super = alloc_set(lc))) {
				dbg_free(name);
				return 0;
			}

			super->name = name;
			super->type = t_raid1;
			super->status = s_setup;
			super->stride = stride(hpt->raid1_shift);
			list_add_tail(&super->list, lc_list(lc, LC_RAID_SETS));
		}

		*super_set = super;
		rs->stride = stride(hpt->raid0_shift);
		if (!find_set(lc, rs->name, 1))
			list_add_tail(&rs->list, &super->sets);

		/* Sort RAID disk into appropriate subset */
		sorted = 0;
		list_for_each_entry(r, &super->sets, list) {
			if (!list_empty(&r->devs))
				h = (struct hpt45x*) list_entry(r->devs.next, typeof(*rd), list)->meta_areas[0].area;

			if (list_empty(&r->devs) ||
			    (h->magic_0 == hpt->magic_0 &&
			     h->raid1_disk_number == hpt->raid1_disk_number)) {
				list_add_dev_sorted(lc, r, rd, _hpt45x_sort);
				sorted++;
				break;
			}
		}

		if (!sorted)
			goto out;

		break;

	case HPT45X_T_SPAN:
	case HPT45X_T_RAID1:
  no_raid10:
		if (!find_set(lc, rs->name, 0))
			list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));

		if (!h || (h->magic_0 == hpt->magic_0)) {
			rs->stride = stride(hpt->raid0_shift);
			list_add_dev_sorted(lc, rs, rd, _hpt45x_sort);
		} else
			goto out;

		break;

	default:
		LOG_ERR(lc, 0, "%s: RAID type %d not supported on %s",
			handler, hpt->type, rd->di->path);
	}

	return 1;

  out:
	log_err(lc, "%s: magic_0 mismatch for %s", handler, rd->di->path);
	return 0;
}

/*
 * Add a Highpoint RAID device to a set.
 */
static struct dmraid_set *hpt45x_group(struct lib_context *lc,
				       struct dmraid_dev *rd)
{
	struct dmraid_set *rs, *super = NULL;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

	/* Unused disk */
	if (!hpt->magic_0)
		return 0;

	if (!(rs = find_set(lc, rd->name, 1))) {
		if (!(rs = alloc_set(lc)))
			return NULL;

		if (!(rs->name = dbg_strdup(rd->name))) {
			free_set(lc, rs);
			return NULL;
		}

		rs->type   = hpt45x_type(lc, rd);
		rs->status = s_setup;
	}

	return group_rd(lc, rs, &super, rd) ? (super ? super : rs) : NULL;
}

/*
 * Read a Highpoint 45X RAID device.
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct hpt45x *hpt)
{
	hpt->magic	= le32_to_cpu(hpt->magic);
	hpt->magic_0	= le32_to_cpu(hpt->magic_0);
	hpt->magic_1	= le32_to_cpu(hpt->magic_1);
	hpt->total_secs	= le32_to_cpu(hpt->total_secs);
}
#endif

/* Magic check. */
static int hpt45x(struct hpt45x *hpt)
{
	return (hpt->magic == HPT45X_MAGIC_OK ||
		hpt->magic == HPT45X_MAGIC_BAD) &&
		hpt->disk_number < 8;
}

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct hpt45x *hpt, struct disk_info *di);
static struct dmraid_dev *hpt45x_read(struct lib_context *lc,
				      struct disk_info *di)
{
	struct hpt45x *hpt;
	struct dmraid_dev *rd = NULL;

	if (!(hpt = alloc_private(lc, handler, sizeof(*hpt))))
		goto out;

	if (!read_file(lc, di->path, (void *) hpt, sizeof(*hpt),
		       HPT45X_CONFIGOFFSET))
		goto bad;

	to_cpu(hpt);
	if (!hpt45x(hpt))
		goto bad;

	file_metadata(lc, handler, di->path, hpt, sizeof(*hpt),
		      HPT45X_CONFIGOFFSET);

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, hpt, di))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(hpt);

   out:
	return rd;
}

/*
 * Write a Highpoint 45X RAID device.
 */
static int hpt45x_write(struct lib_context *lc,
			struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct hpt45x *hpt = rd->meta_areas[0].area;
#endif

	to_disk(hpt);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(hpt);
	return ret;
}

/*
 * Check a Highpoint 45X RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _hpt45x_check(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;

	if (list_empty(&rs->devs))
		return;

	devs = count_devs(lc, rs);
	list_for_each_entry(rd, &rs->devs, list) {
		struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		if (devs != hpt->raid_disks) {
			rd->status = s_broken;
			log_err(lc, "%s: device %s broken in RAID "
			   	"set \"%s\" [%u/%u]",
			   	handler, rd->di->path, rs->name, devs,
				hpt->raid_disks);
		}
	}
}

static int hpt45x_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Check any hierachical RAID sets. */
	check_rs(lc, rs, hpt45x_check);

	_hpt45x_check(lc, rs);

	return set_rs_status(lc, rs) == s_broken ? 0 : 1;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an HPT45X RAID device.
 */
static void hpt45x_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int i;
	struct hpt45x *hpt = (struct hpt45x*) rd->meta_areas[0].area;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("magic: 0x%x", hpt, hpt->magic);
	DP("magic_0: 0x%x", hpt, hpt->magic_0);
	DP("magic_1: 0x%x", hpt, hpt->magic_1);
	DP("total_secs: %u", hpt, hpt->total_secs);
	DP("type: %u", hpt, hpt->type);
	DP("raid_disks: %u", hpt, hpt->raid_disks);
	DP("disk_number: %u", hpt, hpt->disk_number);
	DP("raid0_shift: %u", hpt, hpt->raid0_shift);
	for (i = 0; i < 3; i++)
		DP("dummy[i]: 0x%x", hpt, hpt->dummy[i]);
	DP("raid1_type: %u", hpt, hpt->raid1_type);
	DP("raid1_raid_disks: %u", hpt, hpt->raid1_raid_disks);
	DP("raid1_disk_number: %u", hpt, hpt->raid1_disk_number);
	DP("raid1_shift: %u", hpt, hpt->raid1_shift);

	for (i = 0; i < 3; i++)
		DP("dummy1[i]: 0x%x", hpt, hpt->dummy1[i]);
}
#endif

static struct dmraid_format hpt45x_format = {
	.name	= "hpt45x",
	.descr	= "(+) Highpoint HPT45X",
	.format = FMT_RAID,
	.read	= hpt45x_read,
	.write	= hpt45x_write,
	.group	= hpt45x_group,
	.check	= hpt45x_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= hpt45x_log,
#endif
	.type	= hpt45x_type,
};

/* Register this format handler with the format core */
int register_hpt45x(struct lib_context *lc)
{
	return register_format_handler(lc, &hpt45x_format);
}

/* Set the RAID device contents up derived from the Highpoint ones */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct hpt45x *hpt, struct disk_info *di)
{
	rd->di = di;
	rd->name = _name(lc, hpt, hpt->raid1_type);
	rd->fmt = &hpt45x_format;

	rd->offset = HPT45X_DATAOFFSET;
	rd->sectors = hpt->total_secs;
	rd->sectors += rd->sectors & 1; /* Round up to KBs */
 	rd->sectors /= hpt->raid_disks;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = HPT45X_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*hpt);
	rd->meta_areas[0].area = (void*) hpt;

	rd->status = _status(rd);

	return 1;
}
