/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Intel Software RAID metadata format handler.
 *
 * isw_read() profited from Carl-Daniel Hailfinger's raiddetect code.
 *
 * Profited from the Linux 2.4 iswraid driver by
 * Boji Tony Kannanthanam and Martins Krikis.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "isw.h"
#include <datastruct/byteorder.h>

static const char *handler = "isw";

/*
 * Make up RAID set name from family_num and volume name.
 */
static size_t __name(struct isw *isw, struct isw_dev *dev,
		     char *str, size_t len)
{
	return snprintf(str, len, "isw_%u%s%s", isw->family_num,
			dev ? "_" : "", dev ? (char *) dev->volume : "") + 1;
}

static char *_name(struct lib_context *lc, struct isw *isw, struct isw_dev *dev)
{
        size_t len;
        char *ret;

        len = __name(isw, dev, NULL, 0);
        if ((ret = dbg_malloc(len + 1))) {
                __name(isw, dev, ret, len);
		mk_alpha(ret,
			 strlen(ret) -
			 (dev ? strlen((char *) dev->volume) - 1 : 0));
        } else
                log_err(lc, "%s: failed allocating name string", handler);

        return ret;
}

/* Find a disk table slot by serial number. */
static struct isw_disk *get_disk(struct lib_context *lc,
				 struct isw *isw, struct disk_info *di)
{
	unsigned int i;

	if (!di->serial)
		LOG_ERR(lc, NULL, "%s: no SCSI serial number", handler);

	for (i = 0; i < isw->num_disks; i++) {
		if (!strcmp(di->serial, isw->disk[i].serial))
			return &isw->disk[i];
	}

	return NULL;
}

/*
 * Retrieve status of device.
 *
 * FIXME: is this sufficient to cover all state ?
 */
static enum status __status(unsigned int status)
{
	if ((CONFIGURED_DISK & status) &&
	    (USABLE_DISK & status))
		return s_operational;

	return s_broken;
}

static enum status _status(struct lib_context *lc, struct dmraid_dev *rd)
{
	struct isw *isw = rd->meta_areas[0].area;
	struct isw_disk *disk = get_disk(lc, isw, rd->di);

	if (disk)
		return __status(disk->status);

	LOG_ERR(lc, s_undef, "%s: Error finding disk table slot for %s",
		handler, rd->di->path);
}

/*
 * Mapping of Intel types to generic types.
 */
static struct types types[] = {
        { ISW_T_RAID0, t_raid0},
        { ISW_T_RAID1, t_raid1},
        { 0, t_undef},
};

/* Neutralize disk type. */
static enum type isw_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	struct isw_dev *dev = rd->private;

	if (!dev)
		return t_group;

	return rd_type(types,
		       (unsigned int) dev->vol.map.raid_level);
}

/*
 * Generate checksum of Raid metadata for mpb_size/sizeof(u32) words
 * (checksum field itself ignored for this calculation).
 */
static uint32_t _checksum(struct isw *isw)
{
	uint32_t end = isw->mpb_size / sizeof(end), ret = 0;
	uint32_t *p = (uint32_t*) isw;

	while (end--)
		ret += *p++;

	return ret - isw->check_sum;
}

/* Advance to the next isw_dev from a given one. */
static struct isw_dev *advance_raiddev(struct isw_dev *dev)
{
	int i; /* map size correction */
	struct isw_vol *vol = &dev->vol;
	struct isw_map *map = &vol->map;

	/* Correction */
	i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
	dev++; /* off by i bytes for sure */
	dev = (struct isw_dev *) ((uint8_t *) dev + i);

	if (vol->migr_state) { /* need to add space for another map */
		map = (struct isw_map *) dev; /* Yes, it sits here! */
		i = (map->num_members - 1) * sizeof(map->disk_ord_tbl);
		dev = (struct isw_dev *) ((uint8_t *) dev + sizeof(*map) + i);
	}

	return dev;
}

/* Return isw_dev by table index. */
static struct isw_dev *raiddev(struct isw *isw, unsigned int i)
{
	struct isw_dev *dev = (struct isw_dev*) &isw->disk[isw->num_disks];

	while(i--)
		dev = advance_raiddev(dev);

	return dev;
}

/*
 * Read an Intel RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x, y)
#else
static void to_cpu(struct isw *isw, int full)
{
	unsigned int i, j;
	struct isw_disk *disk;
	struct isw_dev *ird;

	isw->check_sum	    = le32_to_cpu(isw->check_sum);
	isw->mpb_size	    = le32_to_cpu(isw->mpb_size);
	isw->family_num	    = le32_to_cpu(isw->family_num);
	isw->generation_num = le32_to_cpu(isw->generation_num);

	if (!full)
		return;

	for (i = 0; i < isw->num_disks; i++) {
		disk = &isw->disk[i];

		disk->totalBlocks = le32_to_cpu(disk->totalBlocks);
		disk->scsiId	  = le32_to_cpu(disk->scsiId);
		disk->status	  = le32_to_cpu(disk->status);
	}

	for (i = 0; i < isw->num_raid_devs; i++) {
		ird = raiddev(isw, i);

		/* RAID device. */
		ird->SizeLow	     = le32_to_cpu(ird->SizeLow);
		ird->SizeHigh	     = le32_to_cpu(ird->SizeHigh);
		ird->status	     = le32_to_cpu(ird->status);
		ird->reserved_blocks = le32_to_cpu(ird->reserved_blocks);

		/* RAID volume only has 8 bit members. */
	
		/* RAID map. */
		ird->vol.map.pba_of_lba0 = 
			le32_to_cpu(ird->vol.map.pba_of_lba0);
		ird->vol.map.blocks_per_member = 
			le32_to_cpu(ird->vol.map.blocks_per_member);
		ird->vol.map.num_data_stripes = 
			le32_to_cpu(ird->vol.map.num_data_stripes);
		ird->vol.map.blocks_per_strip = 
			le16_to_cpu(ird->vol.map.blocks_per_strip);
		for (j = 0; j < ird->vol.map.num_members; j++) {
			ird->vol.map.disk_ord_tbl[j] = 
				le16_to_cpu(ird->vol.map.disk_ord_tbl[j]);
		}
	}
}
#endif

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct isw *isw, struct disk_info *di, loff_t isw_sboffset);
static struct dmraid_dev *isw_read(struct lib_context *lc, struct disk_info *di)
{
	size_t size = ISW_DISK_BLOCK_SIZE;
	loff_t mpbblocks, isw_sboffset = ISW_CONFIGOFFSET;
	struct isw *isw, *isw_big;
	struct dmraid_dev *rd = NULL;

	if (!(isw = alloc_private(lc, handler, size)))
		goto out;

	if (!read_file(lc, di->path, (void *) isw, size, isw_sboffset)) {
		log_err(lc, "%s: failed reading first sector of ISW superblock "
			"for %s", handler, di->path);
		goto bad;
	}

	/* Check Signature */
	to_cpu(isw, 0);
	if (strncmp(isw->sig, MPB_SIGNATURE, sizeof(MPB_SIGNATURE) - 1))
		goto bad;

	/* Check version info, older versions supported */
	if (strncmp(isw->sig + sizeof(MPB_SIGNATURE) - 1,
		   MPB_VERSION_RAID2, sizeof(MPB_VERSION_RAID2) - 1) > 0) {
		log_print(lc, "%s: superblock found, but version is too high",
			  handler);
		goto bad;
	}

	/* Get the rounded up value for mpbblocks */
	if ((mpbblocks = ((isw->mpb_size + ISW_DISK_BLOCK_SIZE - 1) /
			  ISW_DISK_BLOCK_SIZE)) > 1) {
		size = mpbblocks * ISW_DISK_BLOCK_SIZE;

		/* Allocate memory for extend Intel superblock */
		if (!(isw_big = alloc_private(lc, handler, size)))
			goto bad;
	
		/* Save the first metadata sector */
		memcpy(isw_big, isw, ISW_DISK_BLOCK_SIZE);
		dbg_free(isw);
		isw = isw_big;
	
		/*
		 * Calculate offset to extended Intel superblock
		 * and try to read it.
		 */
		isw_sboffset -= (mpbblocks - 1) * ISW_DISK_BLOCK_SIZE;
		if (!read_file(lc, di->path, (void *) isw + ISW_DISK_BLOCK_SIZE,
			       size - ISW_DISK_BLOCK_SIZE, isw_sboffset)) {
			log_err(lc, "%s: failed reading extended "
				"superblock for %s", handler, di->path);
			goto bad;
		}
	}

	/* Calculate superblock checksum */
	to_cpu(isw, 1);
	if (isw->check_sum != _checksum(isw)) {
		log_err(lc, "%s: extended superblock for %s has wrong checksum",
			handler, di->path);
			goto bad;
	}

	file_metadata(lc, handler, di->path, isw, size, isw_sboffset >> 9);

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, isw, di, isw_sboffset))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(isw);

   out:
	return rd;
}

/*
 * Write an Intel Software RAID device.
 */
static int isw_write(struct lib_context *lc, struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct isw *isw = rd->meta_areas[0].area;
#endif

	to_disk(isw, 1);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(isw, 1);
	return ret;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about an ISW RAID device.
 */
static void isw_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int d, i;
	struct isw *isw = (struct isw*) rd->meta_areas[0].area;
	struct isw_dev *ird;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	P("sig: \"%*s\"", isw, isw->sig, MAX_SIGNATURE_LENGTH, isw->sig);
	DP("check_sum: %u", isw, isw->check_sum);
	DP("mpb_size: %u", isw, isw->mpb_size);
	DP("family_num: %u", isw, isw->family_num);
	DP("generation_num: %u", isw, isw->generation_num);
	DP("reserved[0]: %u", isw, isw->reserved[0]);
	DP("reserved[1]: %u", isw, isw->reserved[1]);
	DP("num_disks: %u", isw, isw->num_disks);
	DP("num_raid_devs: %u", isw, isw->num_raid_devs);
	DP("fill[0]: %u", isw, isw->fill[0]);
	DP("fill[1]: %u", isw, isw->fill[1]);
	for (i = 0; i < ISW_FILLERS; i++) {
		if (isw->filler[i])
        		P("filler[%i]: %u", isw,
			  isw->filler[i], i, isw->filler[i]);
	}

	/* Disk table. */
	for (d = 0; d < isw->num_disks; d++) {
		if (!isw->disk[d].totalBlocks)
			continue;

		P("disk[%u].serial: \"%*s\"", isw,
			  isw->disk[d].serial, d,
			  MAX_RAID_SERIAL_LEN, isw->disk[d].serial);
		P("disk[%u].totalBlocks: %u", isw,
			  isw->disk[d].totalBlocks,
			  d, isw->disk[d].totalBlocks);
		P("disk[%u].scsiId: 0x%x", isw,
			  isw->disk[d].scsiId, d, isw->disk[d].scsiId);
		P("disk[%u].status: 0x%x", isw,
			  isw->disk[d].status, d, isw->disk[d].status);
		for (i = 0; i < ISW_DISK_FILLERS; i++) {
			if (isw->disk[d].filler[i])
                		P("disk[%u].filler[%u]: %u", isw,
				  isw->disk[d].filler[i], d, i,
				  isw->disk[d].filler[i]);
		}
	}

	/* RAID device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		ird = raiddev(isw, d);

		/* RAID device */
		P("isw_dev[%u].volume: \"%*s\"", isw,
		  ird->volume, d, MAX_RAID_SERIAL_LEN, ird->volume);
		P("isw_dev[%u].SizeHigh: %u", isw,
		  ird->SizeHigh, d, ird->SizeHigh);
		P("isw_dev[%u].SizeLow: %u", isw,
		  ird->SizeLow, d, ird->SizeLow);
		P("isw_dev[%u].status: 0x%x", isw,
		  ird->status, d, ird->status);
		P("isw_dev[%u].reserved_blocks: %u", isw,
		  ird->reserved_blocks, d, ird->reserved_blocks);
		for (i = 0; i < ISW_DEV_FILLERS; i++) {
			if (ird->filler[i])
                		P("isw_dev[%u].filler[%u]: %u", isw,
				  ird->filler[i], d, i,
				  ird->filler[i]);
		}

		/* RAID volume */
		for (i = 0; i < 2; i++) {
			if (ird->vol.reserved[i])
                		P("isw_dev[%u].vol.reserved[%u]: %u", isw,
				  ird->vol.reserved[i], d, i,
				  ird->vol.reserved[i]);
		}
		P("isw_dev[%u].vol.migr_state: %u", isw,
		  ird->vol.migr_state, d, ird->vol.migr_state);
		P("isw_dev[%u].vol.migr_type: %u", isw,
		  ird->vol.migr_type, d, ird->vol.migr_type);
		P("isw_dev[%u].vol.dirty: %u", isw,
		  ird->vol.dirty, d, ird->vol.dirty);
		P("isw_dev[%u].vol.fill[0]: %u", isw,
		  ird->vol.fill[0], d, ird->vol.fill[0]);
		for (i = 0; i < 5; i++) {
			if (ird->vol.filler[i])
                		P("isw_dev[%u].vol.filler[%u]: %u", isw,
				  ird->vol.filler[i], d, i,
				  ird->vol.filler[i]);
		}

		/* RAID map */
		P("isw_dev[%u].vol.map.pba_of_lba0: %u", isw,
		  ird->vol.map.pba_of_lba0, d,
		  ird->vol.map.pba_of_lba0);
		P("isw_dev[%u].vol.map.blocks_per_member: %u", isw,
		  ird->vol.map.blocks_per_member, d,
		  ird->vol.map.blocks_per_member);
		P("isw_dev[%u].vol.map.num_data_stripes: %u", isw,
		  ird->vol.map.num_data_stripes, d,
		  ird->vol.map.num_data_stripes);
		P("isw_dev[%u].vol.map.blocks_per_strip: %u", isw,
		  ird->vol.map.blocks_per_strip, d,
		  ird->vol.map.blocks_per_strip);
		P("isw_dev[%u].vol.map.map_state: %u", isw,
		  ird->vol.map.map_state, d,
		  ird->vol.map.map_state);
		P("isw_dev[%u].vol.map.raid_level: %u", isw,
		  ird->vol.map.raid_level, d,
		  ird->vol.map.raid_level);
		P("isw_dev[%u].vol.map.num_members: %u", isw,
		  ird->vol.map.num_members, d,
		  ird->vol.map.num_members);
		for (i = 0; i < 3; i++) {
			if (ird->vol.map.reserved[i])
                		P("isw_dev[%u].vol.map.reserved[%u]: %u", isw,
				  ird->vol.map.reserved[i], d, i,
				  ird->vol.map.reserved[i]);
		}
		for (i = 0; i < 7; i++) {
			if (ird->vol.map.filler[i])
                		P("isw_dev[%u].vol.map.filler[%u]: %u", isw,
				  ird->vol.map.filler[i], d, i,
				  ird->vol.map.filler[i]);
		}
		for (i = 0; i < isw->num_disks; i++) {
			P("isw_dev[%u].vol.map.disk_ord_tbl[%u]: 0x%x", isw,
			  ird->vol.map.disk_ord_tbl[i], d, i,
			  ird->vol.map.disk_ord_tbl[i]);
		}
	}
}
#endif

/*
 * Group an Intel SW RAID disk into potentially
 * multiple RAID sets and RAID disks.
 */
/* Create a RAID device to map a volumes segment. */
static struct dmraid_dev *_create_rd(struct lib_context *lc,
				     struct dmraid_dev *rd,
				     struct isw *isw, struct isw_dev *dev)
{
	struct dmraid_dev *r;

	/* FIXME: FAILED_MAP etc. */
	if (dev->vol.map.map_state) {
		log_print(lc, "%s: unsupported map state 0x%x on %s for %s",
			  handler, dev->vol.map.map_state, rd->di->path,
			  (char *) dev->volume);
		return NULL;
	}

	if (!(r = alloc_dev(lc)))
		return NULL;

	if ((r->private = alloc_private(lc, handler, sizeof*dev)))
		memcpy(r->private, dev, sizeof(*dev));
	else {
		free_dev(lc, r);
		log_err(lc, "%s: alloctaing private memory", handler);
		return NULL;
	}

	r->di = rd->di;
        r->name = _name(lc, isw, dev);
	r->fmt = rd->fmt;
	
	r->offset = dev->vol.map.pba_of_lba0;
	r->sectors = dev->vol.map.blocks_per_member;

	r->status = s_setup;

	return r;
}

/* Find an Intel RAID set or create it. */
static struct dmraid_set *_find_create_rs(struct lib_context *lc,
					  struct list_head *list,
					  struct dmraid_dev *rd,
					  struct isw_dev *dev)
{
	struct dmraid_set *rs;

	if (!(rs = find_set(lc, rd->name, 1))) {
		if (!(rs = alloc_set(lc)))
			return NULL;
	
		if (!(rs->name = dbg_strdup(rd->name))) {
			free_set(lc, rs);
			LOG_ERR(lc, NULL, "%s: allocating RAID set name",
				handler);
		}

		rs->status = s_setup;
		if (dev) {
			rs->type  = isw_type(lc, rd);
			rs->stride = dev->vol.map.blocks_per_strip;
		} else
			rs->type = t_group;

		list_add_tail(&rs->list, list);
	}

	return rs;
}

/*
 * rs_group contains the top-level group RAID set (type: t_group) on entry
 * and shall be returned on success (or NULL on error).
 */
static struct dmraid_set *group_rd(struct lib_context *lc,
				   struct dmraid_set *rs_group,
				   struct dmraid_dev *rd_meta)
{
	unsigned int d;
	struct isw *isw = rd_meta->meta_areas[0].area;
	struct isw_dev *dev;
	struct dmraid_dev *rd;
	struct dmraid_set *rs;

	/* Loop the device/volume table. */
	for (d = 0; d < isw->num_raid_devs; d++) {
		dev = raiddev(isw, d);

		/* FIXME: sufficient ? */
		if (dev->status) {
			log_err(lc, "%s: %s volume is broken",
				handler, dev->volume);
			continue;
		}

		if (!(rd = _create_rd(lc, rd_meta, isw, dev)))
			return NULL;

		if (!(rs = _find_create_rs(lc, &rs_group->sets, rd, dev))) {
			free_dev(lc, rd);
			return NULL;
		}

		list_add_tail(&rd->list, &rs->devs);
	}

	return rs_group;
}

/* Add an Intel SW RAID device to a set */
static struct dmraid_set *isw_group(struct lib_context *lc,
				    struct dmraid_dev *rd_meta)
{
	struct dmraid_set *rs_group;

	/*
	 * Once we get here, an Intel SW RAID disk containing a metadata area
	 * with a volume table has been discovered by isw_read.
	 */
	/* Check if a top level group RAID set already exists. */
	if (!(rs_group = _find_create_rs(lc, lc_list(lc, LC_RAID_SETS),
					 rd_meta, NULL)))
		return NULL;

	/* Add the whole underlying (meta) RAID device to the group set. */
	list_add_tail(&rd_meta->list, &rs_group->devs);

	/*
	 * We need to run through the volume table and create a RAID set and
	 * RAID devices hanging off it for every volume,
	 * so that the activate code is happy.
	 *
	 * A pointer to the top-level group RAID set
	 * gets returned or NULL on error.
	 */
	return group_rd(lc, rs_group, rd_meta);
}

/*
 * Check an Intel SW RAID set.
 *
 * FIXME: more sanity checks.
 */
static void __isw_check(struct lib_context *lc, struct dmraid_set *rs)
{
	struct dmraid_dev *rd;

	if (rs->type == t_group)
		return;

	list_for_each_entry(rd, &rs->devs, list) {
		struct isw_dev *dev = (struct isw_dev *) rd->private;

		/* FIXME: more status checks ? */
		rd->status = dev->status ? s_broken : s_operational;
		if (rd->status == s_broken)
			log_err(lc, "%s device for volume \"%s\" broken on %s "
				"in RAID set \"%s\"",
				handler, dev->volume, rd->di->path, rs->name);

		log_dbg(lc, "set status of %s device on \"%s\" to %u", handler,
			rd->di->path, rd->status);
	}
}

static int _isw_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Check any hierachical RAID sets. */
	check_rs(lc, rs, _isw_check);

	__isw_check(lc, rs);

	return set_rs_status(lc, rs) == s_broken ? 0 : 1;
}

static int isw_check(struct lib_context *lc, struct dmraid_set *rs_group)
{
	if (rs_group->type == t_group)
		return _isw_check(lc, rs_group);

	return 0;
}

static struct dmraid_format isw_format = {
	.name	= "isw",
	.descr	= "(+) Intel Software RAID",
	.format = FMT_RAID,
	.read	= isw_read,
	.write	= isw_write,
	.group	= isw_group,
	.check	= isw_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= isw_log,
#endif
	.type	= isw_type,
};

/* Register this format handler with the format core */
int register_isw(struct lib_context *lc)
{
	return register_format_handler(lc, &isw_format);
}

/*
 * Set the RAID device contents up derived from the Intel ones.
 *
 * This is the first one we get with here and we potentially need to
 * create many in isw_group() in case of multiple Intel SW RAID devices
 * on this RAID disk.
 */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct isw *isw, struct disk_info *di, loff_t isw_sboffset)
{
	rd->di = di;
        rd->name = _name(lc, isw, NULL);
	rd->fmt = &isw_format;

	rd->offset = ISW_DATAOFFSET;
	rd->sectors = isw_sboffset / ISW_DISK_BLOCK_SIZE; /* FIXME: Correct ? */

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = rd->sectors;
	rd->meta_areas[0].size = di->sectors - rd->sectors - 1;
	rd->meta_areas[0].area = (void*) isw;

	rd->status = _status(lc, rd);

	return 1;
}
