/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * LSI Logic MegaRAID (and MegaIDE ?) ATARAID metadata format handler.
 *
 * Needs more metadata reengineering and grouping logic coding.
 *
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "lsi.h"
#include <datastruct/byteorder.h>

static const char *handler = "lsi";

/* Make up RAID device name. */
/* FIXME: senseful name ;) */
static size_t __name(struct lsi *lsi, char *str, size_t len)
{
	return snprintf(str, len, "lsi_%d", lsi->set_number) + 1;
}

static char *_name(struct lib_context *lc, struct lsi *lsi, int type)
{
	size_t len;
	char *ret;

	len = __name(lsi, NULL, 0);
	if ((ret = dbg_malloc(len + 1))) {
		__name(lsi, ret, len);
		mk_alpha(ret, strlen(ret));
	} else
                log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/* Mapping of LSI Logic types to generic types */
static struct types types[] = {
	{ LSI_MAGIC_T_RAID0, t_raid0 },
	{ LSI_MAGIC_T_RAID1, t_raid1 },
	{ LSI_MAGIC_T_RAID10, t_raid1 },
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type lsi_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type(types,
		       (unsigned int) ((struct lsi*) rd->meta_areas[0].area)->type);
}

/*
 * Decide about ordering sequence of RAID device.
 * Called by list_add_dev_sorted().
 */
static int _lsi_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct lsi*) new->meta_areas[0].area)->disk_number <
	       ((struct lsi*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Read an LSI Logic RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct lsi *lsi)
{
	struct lsi_disks *dsks;

	lsi->stride = le16_to_cpu(lsi->stride);

	for (dsks = lsi->disks; dsks < &lsi->disks[4]; dsks++) {
		dsks->magic_0 = le16_to_cpu(dsks->magic_0);
		dsks->magic_1 = le16_to_cpu(dsks->magic_1);
	}

	lsi->what = le32_to_cpu(lsi->what);
}
#endif

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct lsi *lsi, struct disk_info *di);
static struct dmraid_dev *lsi_read(struct lib_context *lc, struct disk_info *di)
{
	struct lsi *lsi;
	struct dmraid_dev *rd = NULL;

	if (!(lsi = alloc_private(lc, handler, sizeof(*lsi))))
		goto out;

	if (!read_file(lc, di->path, (void *) lsi, sizeof(*lsi),
		       LSI_CONFIGOFFSET))
		goto bad;

	to_cpu(lsi);
	if (strncmp(lsi->magic_name, LSI_MAGIC_NAME,
		    sizeof(LSI_MAGIC_NAME) - 1))
		goto bad;

	file_metadata(lc, handler, di->path, lsi, sizeof(*lsi),
		      LSI_CONFIGOFFSET);

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, lsi, di))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(lsi);

   out:
	return rd;
}

/*
 * Write a LSI Logic RAID device.
 */
static int lsi_write(struct lib_context *lc, struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct lsi *lsi = rd->meta_areas[0].area;
#endif

	to_disk(lsi);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(lsi);
	return ret;
}

/*
 * Group the RAID disk into a set.
 *
 * FIXME: this needs more work together with the metadata reengineering.
 */
static int group_rd(struct lib_context *lc, struct dmraid_set *rs,
		    struct dmraid_dev *rd)
{
	struct lsi *lsi = (struct lsi*) rd->meta_areas[0].area;

	if ((rs->type = rd_type(types, lsi->type)) == t_undef)
		LOG_ERR(lc, 0, "%s: RAID type %u not supported",
			handler, lsi->type);

	rs->stride = lsi->stride;

	switch (lsi->type) {
	case LSI_MAGIC_T_RAID0:
	case LSI_MAGIC_T_RAID1:
		list_add_dev_sorted(lc, rs, rd, _lsi_sort);
		break;

	case LSI_MAGIC_T_RAID10:
		LOG_ERR(lc, 0, "%s: RAID10 not supported yet", handler);
	}

	return 1;
}

static struct dmraid_set *lsi_group(struct lib_context *lc,
				    struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	if (!(rs = find_set(lc, rd->name, 0))) {
		if (!(rs = alloc_set(lc)))
			return NULL;

		if (!(rs->name = dbg_strdup(rd->name))) {
			free_set(lc, rs);
			return NULL;
		}

		rs->type = lsi_type(lc, rd);
		rs->status = s_setup;
		list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));
	}

	return group_rd(lc, rs, rd) ? rs : NULL;
}

/*
 * Check an LSI RAID set.
 *
 * FIXME: add checks!!!
 */
static int lsi_check(struct lib_context *lc, struct dmraid_set *rs)
{
	return 1;
}

#ifdef DMRAID_NATIVE_LOG
/* Log native information about an LSI Logic RAID device. */
static void lsi_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int i;
	struct lsi *lsi = (struct lsi*) rd->meta_areas[0].area;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("magic_name: \"%s\"", lsi, lsi->magic_name);
	DP("seqno: %u", lsi, lsi->seqno);
	DP("type: %u", lsi, lsi->type);
	DP("stride: %u", lsi, lsi->stride);

	for (i = 0; i < 4; i++) {
		P("disks[%u].unknown: %u, 0x%x", lsi, lsi->disks[i].unknown, i,
		  lsi->disks[i].unknown, lsi->disks[i].unknown);
		P("disks[%u].magic_0: 0x%x, %x, %x", lsi,
		  lsi->disks[i].magic_0, i,
		  lsi->disks[i].magic_0,
		  (unsigned char) (((char *) &lsi->disks[i].magic_0)[0]),
		  (unsigned char) (((char *) &lsi->disks[i].magic_0)[1]));
		P("disks[%u].magic_1: 0x%x, %x, %x", lsi,
		  lsi->disks[i].magic_1, i,
		  lsi->disks[i].magic_1,
		  (unsigned char) (((char *) &lsi->disks[i].magic_1)[0]),
		  (unsigned char) (((char *) &lsi->disks[i].magic_1)[1]));
		P("disks[%u].disk_number: %u", lsi, lsi->disks[i].disk_number,
		  i, lsi->disks[i].disk_number);
		P("disks[%u].set_number: %u", lsi, lsi->disks[i].set_number,
		  i, lsi->disks[i].set_number);
	}
	DP("disk_number: %u", lsi, lsi->disk_number);
	DP("set_number: %u", lsi, lsi->set_number);
	P("what: %u, 0x%x", lsi, lsi->what, lsi->what, lsi->what);
}
#endif

static struct dmraid_format lsi_format = {
	.name	= "lsi",
	.descr	= "(0) LSI Logic MegaRAID",
	.format = FMT_RAID,
	.read	= lsi_read,
	.write	= lsi_write,
	.group	= lsi_group,
	.check	= lsi_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= lsi_log,
#endif
	.type	= lsi_type,
};

/* Register this format handler with the format core */
int register_lsi(struct lib_context *lc)
{
	return register_format_handler(lc, &lsi_format);
}

/* Set the RAID device contents up derived from the LSI ones */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct lsi *lsi, struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(lc, lsi, 0);
	rd->fmt = &lsi_format;

	rd->offset = LSI_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* FIXME: propper size */
	rd->sectors -= rd->sectors & 1;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = LSI_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*lsi);
	rd->meta_areas[0].area = (void*) lsi;

	rd->status = s_operational; /* FIXME: derive from metadata. */

	return 1;
}
