/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Promise FastTrak ATARAID metadata format handler.
 *
 * pdc_read(), pdc_group() and group_rd() profited from
 * Carl-Daniel Hailfinger's raiddetect code.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "pdc.h"
#include <datastruct/byteorder.h>

static const char *handler = "pdc";

/* Make up RAID device name. */
/* FIXME: better name ? */
static size_t __name(struct pdc *pdc, char *str, size_t len)
{
	return snprintf(str, len, "pdc_%u", pdc->magic_1) + 1;
}

static char *_name(struct lib_context *lc, struct pdc *pdc, int type)
{
	size_t len;
	char *ret = NULL;

	len = __name(pdc, ret, 0);
	if ((ret = dbg_malloc(len + 1))) {
		__name(pdc, ret, len);
		mk_alpha(ret, strlen(ret));
	} else
                log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status _status(struct dmraid_dev *rd)
{
	return (((struct pdc*) rd->meta_areas[0].area)->raid.flags & 0x80) ?
	       s_operational : s_broken; /* FIXME: */
}

/* Mapping of Promise types to generic types */
static struct types types[] = {
        { PDC_T_SPAN,   t_linear},
        { PDC_T_RAID0, t_raid0},
        { PDC_T_RAID1, t_raid1},
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type pdc_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type(types,
		       (unsigned int) ((struct pdc*) rd->meta_areas[0].area)->raid.type);
}

/* Calculate checksum on Promise metadata. */
static uint32_t _checksum(struct pdc *pdc)
{
	unsigned int i = 511, sum = 0;
	uint32_t *p = (uint32_t*) pdc;

	while (i--)
		sum += *p++;

	return sum == pdc->checksum;
}

/* Calculate metadata offset. */
static loff_t _offset(uint64_t size, unsigned int sectors)
{
	return (size - sectors) << 9;
}

/*
 * Read a Promise FastTrak RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct pdc *pdc)
{
	struct pdc_disk *dsk;

	pdc->magic_0		= le32_to_cpu(pdc->magic_0);
	pdc->magic_1		= le32_to_cpu(pdc->magic_1);
	pdc->raid.flags		= le32_to_cpu(pdc->raid.flags);
	pdc->raid.magic_0	= le32_to_cpu(pdc->raid.magic_0);
	pdc->raid.disk_secs	= le32_to_cpu(pdc->raid.disk_secs);
	pdc->raid.total_secs	= le32_to_cpu(pdc->raid.total_secs);
	pdc->raid.cylinders	= le16_to_cpu(pdc->raid.cylinders);
	pdc->raid.magic_1	= le32_to_cpu(pdc->raid.magic_1);

	for (dsk = pdc->raid.disk; dsk < &pdc->raid.disk[8]; dsk++) {
		dsk->magic_0	 = le32_to_cpu(dsk->magic_0);
		dsk->disk_number = le32_to_cpu(dsk->disk_number);
	}
}
#endif

/* Magic check. */
static int is_pdc(struct pdc *pdc)
{
	return pdc->magic_0 == pdc->raid.magic_0 &&
// FIXME: Gerte Hoogewerf RAID0 metadata doesn't have magic_1's equal ?
//	       pdc->magic_1 == pdc->raid.magic_1 &&
	       pdc->raid.total_disks < 8;
}

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct pdc *pdc, struct disk_info *di,
		    unsigned int meta_sector);
static struct dmraid_dev *pdc_read(struct lib_context *lc, struct disk_info *di)
{ 
	static unsigned int *sector, sectors[] = {
		PDC_CONFIGOFFSET,
		255,
		256,
		0
	};
	struct pdc *pdc;
	struct dmraid_dev *rd = NULL;

	if (!(pdc = alloc_private(lc, handler, sizeof(*pdc))))
		goto out;

	/* Assume certain tracks lengths. */
	for (sector = sectors; *sector; sector++) {
		if (!read_file(lc, di->path, (void *) pdc, sizeof(*pdc),
			       _offset(di->sectors, *sector)))
			goto bad;

        	if (!strncmp(pdc->promise_id, PDC_MAGIC,
			     sizeof(PDC_MAGIC) - 1))
			break;
	}

	to_cpu(pdc);
	if (!*sector)
		goto bad;

	if (!is_pdc(pdc)) {
		log_err(lc, "%s: identifying %s, magic_0: 0x%x/0x%x, "
			"magic_1: 0x%x/0x%x, total_disks: %u",
			handler, di->path,
			pdc->magic_0, pdc->raid.magic_0,
			pdc->magic_1, pdc->raid.magic_1, pdc->raid.total_disks);
		goto bad;
	}

	if (!_checksum(pdc)) {
		log_err(lc, "%s: invalid checksum on %s", handler, di->path);
		goto bad;
	}

	file_metadata(lc, handler, di->path, pdc, sizeof(*pdc),
		      _offset(di->sectors, *sector));

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, pdc, di, *sector))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(pdc);

   out:
	return rd;
}

/*
 * Write a Promise FastTrak RAID device.
 */
static int pdc_write(struct lib_context *lc, struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct pdc *pdc = rd->meta_areas[0].area;
#endif

	to_disk(pdc);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(pdc);
	return ret;
}

/*
 * Decide about ordering sequence of RAID device.
 * (Called by list_add_dev_sorted().
 */
static int pdc_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct pdc*) new->meta_areas[0].area)->raid.disk_number <
	       ((struct pdc*) rd->meta_areas[0].area)->raid.disk_number;
}

/*
 * Group the RAID disk into a Promise set.
 *
 * Check device hierarchy and create super set appropriately.
 */
static unsigned int stride(struct pdc *pdc)
{
	return pdc->raid.raid0_shift ? 1 << pdc->raid.raid0_shift : 0;
}

static int group_rd(struct lib_context *lc,
		    struct dmraid_set *rs, struct dmraid_dev *rd)
{
	struct pdc *pdc = (struct pdc*) rd->meta_areas[0].area;

	switch (pdc->raid.type) {
	case PDC_T_SPAN:
	case PDC_T_RAID0:
	case PDC_T_RAID1:
			if (!find_set(lc, rs->name, 0))
			list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));

		list_add_dev_sorted(lc, rs, rd, pdc_sort);
		break;

	default:
		log_err(lc, "%s: invalid type for %s", rd->di->path, handler);
		return 0;
	}

	return 1;
}

/* Add a PDC RAID device to a set. */
static struct dmraid_set *pdc_group(struct lib_context *lc,
				    struct dmraid_dev *rd)
{
	struct dmraid_set *rs;
	struct pdc *pdc = (struct pdc*) rd->meta_areas[0].area;

	if (!(rs = find_set(lc, rd->name, 1))) {
		if (!(rs = alloc_set(lc)))
			return NULL;

		if (!(rs->name = dbg_strdup(rd->name))) {
			free_set(lc, rs);
			LOG_ERR(lc, NULL, "%s: allocating RAID set name", handler);
		}

		rs->type   = pdc_type(lc, rd);
		rs->status = s_setup;
		rs->stride = stride(pdc);
	}

	return group_rd(lc, rs, rd) ? rs : NULL;
}

/*
 * Check a PDC RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _pdc_check(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;

	if (list_empty(&rs->devs))
		return;

	devs = count_devs(lc, rs);
	list_for_each_entry(rd, &rs->devs, list) {
		unsigned char disks;
		struct pdc *pdc = (struct pdc*) rd->meta_areas[0].area;

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		switch (pdc->raid.type) {
		case PDC_T_SPAN:
		case PDC_T_RAID0:
		case PDC_T_RAID1:
			disks = pdc->raid.total_disks;
			break;

		default:
			disks = 0;
		}

		if (devs != disks) {
			rd->status = s_broken;
			log_err(lc, "%s: device %s broken in RAID "
				"set \"%s\" [%u/%u]",
				handler, rd->di->path, rs->name, devs, disks);
		}
	}
}

static int pdc_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Check hierachical RAID sets. */
	check_rs(lc, rs, pdc_check);

	_pdc_check(lc, rs);

	return set_rs_status(lc, rs) == s_broken ? 0 : 1;
}

#ifdef DMRAID_NATIVE_LOG
/* Log native information about a Promise RAID device. */
static void pdc_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	unsigned int i;
	struct pdc *pdc = (struct pdc*) rd->meta_areas[0].area;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("promise_id: \"%s\"", pdc, pdc->promise_id);
	DP("unknown_0: 0x%x", pdc, pdc->unknown_0);
	DP("magic_0: 0x%x", pdc, pdc->magic_0);
	DP("unknown_1: 0x%x", pdc, pdc->unknown_1);
	DP("magic_1: 0x%x", pdc, pdc->magic_1);
	DP("unknown_2: 0x%x", pdc, pdc->unknown_2);

	DP("raid.flags: 0x%x", pdc, pdc->raid.flags);
	DP("raid.unknown_0: 0x%x", pdc, pdc->raid.unknown_0);
	DP("raid.disk_number: %u", pdc, pdc->raid.disk_number);
	DP("raid.channel: %u", pdc, pdc->raid.channel);
	DP("raid.device: %u", pdc, pdc->raid.device);
	DP("raid.magic_0: 0x%x", pdc, pdc->raid.magic_0);
	DP("raid.unknown_1: 0x%x", pdc, pdc->raid.unknown_1);
	DP("raid.unknown_2: 0x%x", pdc, pdc->raid.unknown_2);
	DP("raid.disk_secs: %u", pdc, pdc->raid.disk_secs);
	DP("raid.unknown_3: 0x%x", pdc, pdc->raid.unknown_3);
	DP("raid.unknown_4: 0x%x", pdc, pdc->raid.unknown_4);
	DP("raid.status: 0x%x", pdc, pdc->raid.status);
	DP("raid.type: 0x%x", pdc, pdc->raid.type);
	DP("raid.total_disks: %u", pdc, pdc->raid.total_disks);
	DP("raid.raid0_shift: %u", pdc, pdc->raid.raid0_shift);
	DP("raid.raid0_disks: %u", pdc, pdc->raid.raid0_disks);
	DP("raid.array_number: %u", pdc, pdc->raid.array_number);
	DP("raid.total_secs: %u", pdc, pdc->raid.total_secs);
	DP("raid.cylinders: %u", pdc, pdc->raid.cylinders);
	DP("raid.heads: %u", pdc, pdc->raid.heads);
	DP("raid.sectors: %u", pdc, pdc->raid.sectors);
	DP("raid.magic_1: 0x%x", pdc, pdc->raid.magic_1);
	DP("raid.unknown_5: 0x%x", pdc, pdc->raid.unknown_5);

	for (i = 0; i < 8; i++) {
		if (!pdc->raid.disk[i].magic_0)
			continue;

		P2("raid.disk[%d].unknown_0: 0x%x", pdc,
		   i, pdc->raid.disk[i].unknown_0);
		P2("raid.disk[%d].channel: %u", pdc,
		   i, pdc->raid.disk[i].channel);
		P2("raid.disk[%d].device: %u", pdc,
		   i, pdc->raid.disk[i].device);
		P2("raid.disk[%d].magic_0: 0x%x", pdc,
		   i, pdc->raid.disk[i].magic_0);
		P2("raid.disk[%d].disk_number: %u", pdc,
		   i, pdc->raid.disk[i].disk_number);
	}

	P("checksum: 0x%x %s", pdc, pdc->checksum, pdc->checksum,
	   _checksum(pdc) ? "Ok" : "Crap");
}
#endif

static struct dmraid_format pdc_format = {
	.name	= "pdc",
	.descr	= "(+) Promise FastTrack",
	.format = FMT_RAID,
	.read	= pdc_read,
	.write	= pdc_write,
	.group	= pdc_group,
	.check	= pdc_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= pdc_log,
#endif
	.type	= pdc_type,
};

/* Register this format handler with the format core */
int register_pdc(struct lib_context *lc)
{
	return register_format_handler(lc, &pdc_format);
}

/* Set the RAID device contents up derived from the PDC ones */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct pdc *pdc, struct disk_info *di,
		    unsigned int meta_sector)
{
        rd->di = di;
        rd->name = _name(lc, pdc, 0);
	rd->fmt = &pdc_format;
	rd->offset = PDC_DATAOFFSET;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = di->sectors - meta_sector;
	rd->meta_areas[0].size = sizeof(*pdc);
	rd->meta_areas[0].area = (void*) pdc;

	rd->status = _status(rd);

	switch (pdc->raid.type) {
	case PDC_T_SPAN:
		rd->sectors = di->sectors - meta_sector;
		break;

	case PDC_T_RAID0:
	case PDC_T_RAID1:
		rd->sectors = pdc->raid.total_secs / pdc->raid.total_disks;
		break;

	default:
		log_err(lc, "%s: type 0x%x not supported", handler, pdc->raid.type);
	}

	return 1;
}
