/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Silicon Image Medley ATARAID metadata format handler.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "sil.h"
#include <datastruct/byteorder.h>

static const char *handler = "sil";

/* Make up RAID device name from some 'magic' numbers */
/* FIXME: better name ? */
static size_t __name(struct sil *sil, char *str, size_t len,
		     unsigned char type)
{
	const char *fmt = type ? "sil_%02u%02u%02u%02u%02u%02u%u-%u" :
				 "sil_%02u%02u%02u%02u%02u%02u%u";

	return snprintf(str, len, fmt,
			sil->year, sil->month, sil->day,
			sil->hour, sil->minutes % 60, sil->seconds % 60,
			sil->incarnation_no, sil->mirrored_set_number) + 1;
}

static char *_name(struct lib_context *lc, struct sil *sil, unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(sil, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1)))
		__name(sil, ret, len, type);
	else
		log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/*
 * Retrieve status of device.
 * FIXME: is this sufficient to cover all state ?
 */
static enum status _status(struct dmraid_dev *rd)
{
        struct sil *sil = (struct sil*) rd->meta_areas[0].area;

	return sil->member_status == 1 ? s_operational : s_broken;
}

/* Mapping of SIL 680 types to generic types */
static struct types types[] = {
        { SIL_T_RAID0,  t_raid0},
        { SIL_T_RAID1,  t_raid1},
        { SIL_T_RAID10, t_raid0},
        { SIL_T_SPARE,   t_spare},
        { 0, t_undef}
};

/* Neutralize disk type */
static enum type sil_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type (types,
			(unsigned int) ((struct sil*) rd->meta_areas[0].area)->type);
}

/* Calculate checksum on metadata */
static uint16_t _checksum(struct sil *sil, unsigned int count)
{
	uint16_t *p = (uint16_t*) sil, sum = 0;

        while (count--)
		sum += *p++;

	return sum;
}

/* Calculate all checksums on metadata */
/* FIXME: checksum1 check correct ? */
static uint32_t checksum(struct sil* sil) {
	if (_checksum(sil, 160) ||
	    _checksum((struct sil*) &sil->auto_rebuild,
		      (unsigned int) &sil->checksum1 -
		      (unsigned int) &sil->auto_rebuild) != sil->checksum1 ||
	    _checksum(sil, 256) != sil->checksum2)
		return 0;

	return 1;
}

/*
 * Read a Silicon Image RAID device
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct sil *sil)
{
	sil->magic		= le32_to_cpu(sil->magic);
	sil->array_sectors_low	= le32_to_cpu(sil->array_sectors_low);
	sil->array_sectors_high = le32_to_cpu(sil->array_sectors_high);
	sil->thisdisk_sectors	= le32_to_cpu(sil->thisdisk_sectors);
	sil->product_id		= le16_to_cpu(sil->product_id);
	sil->vendor_id		= le16_to_cpu(sil->vendor_id);
	sil->minor_ver		= le16_to_cpu(sil->minor_ver);
	sil->major_ver		= le16_to_cpu(sil->major_ver);
	sil->raid0_stride	= le16_to_cpu(sil->raid0_stride);
	sil->rebuild_ptr_low	= le32_to_cpu(sil->rebuild_ptr_low);
	sil->rebuild_ptr_high	= le32_to_cpu(sil->rebuild_ptr_high);
	sil->incarnation_no	= le32_to_cpu(sil->incarnation_no);
	sil->checksum1		= le16_to_cpu(sil->checksum1);
	sil->checksum2		= le16_to_cpu(sil->checksum2);
}
#endif

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct sil *sil, struct disk_info *di);
#define	AREAS	4
#define SIL_META_AREA(i)	(SIL_CONFIGOFFSET - (i * 512 << 9))
static int _file_name(char *str, size_t len, char *name, int i)
{
	return snprintf(str, len, "%s_%d", name, i) + 1;
}

static char *file_name(struct lib_context *lc, char *name, int i)
{
	size_t len = _file_name(NULL, 0, name, i);
	char *str;

	if ((str = dbg_malloc(len))) {
		_file_name(str, len, name, i);
		return str;
	}

	LOG_ERR(lc, NULL, "%s: allocating file name", handler);
}

static struct dmraid_dev *sil_read(struct lib_context *lc, struct disk_info *di)
{
	unsigned int areas = 0, i;
	struct sil *sil, *sils[AREAS];
	struct dmraid_dev *rd = NULL;

	memset(sils, 0, sizeof(sils));
	for (i = 0; i < AREAS; i++) {
		if (!(sils[i] = alloc_private(lc, handler, sizeof(*sil))))
			goto bad;
	}

	/* Read all metadata areas. */
	for (i = 0; i < AREAS; i++) {
		if (!read_file(lc, di->path, (void *) sils[i], sizeof(*sils[i]),
			       SIL_META_AREA(i)))
			goto bad;

		to_cpu(sils[i]);
		if (sils[i]->magic == SIL_MAGIC)
			areas++;
	}

	/* No SIL metadata areas found. */
	if (!areas)
		goto bad;

	for (i = 0; i < AREAS; i++) {
		char *name;

		if (sils[i]) {
			if (!(name = file_name(lc, di->path, i)))
				goto bad;

			file_metadata(lc, handler, name, sils[i],
				      sizeof(*sil), SIL_META_AREA(i) >> 9);
			dbg_free(name);
		}
	}

	sil = sils[0];
	for (i = 1; i < AREAS; i++) {
		if (sils[i]) {
			if (memcmp(sil, sils[i], sizeof(*sil))) {
				log_err(lc, "%s: metadata copy %u differs",
					handler, i);
				goto bad;
			}
		} else
			log_err(lc, "%s: metadata area %d missing", handler, i);
	}

	/* Some metadata areas missing! */
	if (areas != AREAS) {
		/*
		 * FIXME: update any missing metadata areas
		 * (with voting) automatically ?
		 */
		log_err(lc, "%s: only %u/%u metadata areas found on %s",
			handler, areas, AREAS, di->path);
		goto bad;
	}

	if (sil->major_ver != 2)
		log_warn(lc, "%s: major version %u; format handler "
			  "tested on version 2 only", handler, sil->major_ver);

	if (!checksum(sil)) {
		log_err(lc, "%s: invalid metadata checksum on %s",
			handler, di->path);
		goto bad;
	}

	if (di->sectors < sil->thisdisk_sectors) {
		log_err(lc, "%s: invalid disk size in metadata on %s",
			handler, di->path);
		goto bad;
	}

	if (!(rd = alloc_dev(lc)))
		goto bad;

	if (setup_rd(lc, rd, sil, di)) {
		i = 1;
		goto out;
	}

	free_dev(lc, rd);
	rd = NULL;

  bad:
	i = 0;

  out:
	for (; i < AREAS; i++) {
		if (sils[i])
			dbg_free(sils[i]);
		else
			break;
	}

	return rd;
}

/*
 * Write a Silicon Image RAID device.
 */
static int sil_write(struct lib_context *lc, struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct sil *sil = rd->meta_areas[0].area;
#endif

	to_disk(sil);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(sil);
	return ret;
}

/*
 * Decide about ordering sequence of RAID device.
 * (Called by list_add_dev_sorted().
 */
static int _sil_sort(struct dmraid_dev *rd, struct dmraid_dev *new)
{
	return ((struct sil*) new->meta_areas[0].area)->disk_number <
	       ((struct sil*) rd->meta_areas[0].area)->disk_number;
}

/*
 * Group the RAID disk into a SIL set.
 *
 * Check device hierarchy and create super set appropriately.
 */
static int group_rd(struct lib_context *lc, struct dmraid_set *rs,
		    struct dmraid_set **super_set, struct dmraid_dev *rd)
{
	unsigned int sorted;
	char *name;
	struct sil *s, *sil = (struct sil*) rd->meta_areas[0].area;
	struct dmraid_set *r, *super;

	switch (sil->type) {
	case SIL_T_RAID0:
	case SIL_T_RAID1:
		if (!(find_set(lc, rs->name, 0)))
			list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));

		list_add_dev_sorted(lc, rs, rd, _sil_sort);
		break;

	case SIL_T_RAID10:
		/*
		 * We've got a striped raid set with a mirror on top
		 * when we get here.
		 * Let's find and optionally allocate the mirror set on top.
		 */

		/* Superset name */
		if (!(name = _name(lc, sil, 0)))
			return 0;

		if ((super = find_set(lc, name, 0)))
			dbg_free(name);
		else {
			if (!(super = alloc_set(lc))) {
				dbg_free(name);
				return 0;
			}

			super->name = name;
			super->type = t_raid1;
			super->status = s_setup;
			super->stride = sil->raid0_stride;
			list_add_tail(&super->list, lc_list(lc, LC_RAID_SETS));
		}

		*super_set = super;
		rs->stride = sil->raid0_stride;
		if (!(find_set(lc, rs->name, 1)))
			list_add_tail(&rs->list, &super->sets);

		/* Sort RAID disk into appropriate subset */
		sorted = 0;
		list_for_each_entry(r, &super->sets, list) {
			if (!list_empty(&r->devs))
				s = (struct sil*) list_entry(r->devs.next, typeof(*rd), list)->meta_areas[0].area;

			if (list_empty(&r->devs) ||
			    (sil->mirrored_set_number ==
			     s->mirrored_set_number)) {
				list_add_dev_sorted(lc, r, rd, _sil_sort);
				sorted++;
				break;
			}
		}

		if (!sorted)
			goto out;

		break;

	case SIL_T_SPARE:
		break;

	default:
		return 0;
	}

	return 1;

  out:
	log_err(lc, "%s: mismatch for %s", handler, rd->di->path);
	return 0;
}

/* Add a SIL RAID device to a set */
static struct dmraid_set *sil_group(struct lib_context *lc,
				    struct dmraid_dev *rd)
{
	struct dmraid_set *rs, *super = NULL;
	struct sil *sil = (struct sil*) rd->meta_areas[0].area;

	if (sil->disk_number > 7)
		return 0;
	
	if (!(rs = find_set(lc, rd->name, 1))) {
		if (!(rs = alloc_set(lc)))
			return NULL;

		if (!(rs->name = dbg_strdup(rd->name))) {
			free_set(lc, rs);
			LOG_ERR(lc, NULL, "%s: allocating RAID set name",
				handler);
		}

		rs->type   = sil_type(lc, rd);
		rs->status = s_setup;
		rs->stride = sil->raid0_stride;
	}

	return group_rd(lc, rs, &super, rd) ? (super ? super : rs) : NULL;
}

/*
 * Check a SIL RAID set.
 *
 * FIXME: more sanity checks.
 */
static void _sil_check(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int devs;
	struct dmraid_dev *rd;

	if (list_empty(&rs->devs))
		return;

	devs = count_devs(lc, rs);
	list_for_each_entry(rd, &rs->devs, list) {
		unsigned char disks;
		struct sil *sil = (struct sil*) rd->meta_areas[0].area;

		log_dbg(lc, "checking %s device \"%s\"", handler, rd->di->path);
		switch (sil->type) {
		case SIL_T_RAID0:
		case SIL_T_RAID10:
			disks = sil->drives_per_striped_set;
			break;

		case SIL_T_RAID1:
			disks = sil->drives_per_mirrored_set;
			break;

		default:
			disks = 0;
		}

		if (devs != disks) {
			rd->status = s_inconsistent;
			log_err(lc, "%s: wrong # of devices in RAID "
				"set \"%s\" [%u/%u]",
				handler, rs->name, devs, disks);
		}
	}
}

static int sil_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Check hierachical RAID sets */
	check_rs(lc, rs, sil_check);

	_sil_check(lc, rs);

	return set_rs_status(lc, rs) == s_broken ? 0 : 1;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about a Silicon Image  RAID device.
 */
static void sil_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	struct sil *sil = (struct sil*) rd->meta_areas[0].area;

	log_print(lc, "%s (%s):", rd->di->path, handler);
	DP("unknown0: \"%42s\"", sil, sil->unknown0);
	DP("ascii_version: \"%8s\"", sil, sil->ascii_version);
	DP("diskname: \"%32s\"", sil, sil->diskname);
	DP("unknown1: \"%22s\"", sil, sil->unknown1);
	DP("magic: 0x%x", sil, sil->magic);
	DP("unknown1a: \"%8s\"", sil, sil->unknown1a);
	DP("array_sectors_low: %u", sil, sil->array_sectors_low);
	DP("array_sectors_high: %u", sil, sil->array_sectors_high);
	DP("unknown2: \"%4s\"", sil, sil->unknown2);
	DP("thisdisk_sectors: %u", sil, sil->thisdisk_sectors);
	DP("product_id: %u", sil, sil->product_id);
	DP("vendor_id: %u", sil, sil->vendor_id);
	DP("minor_ver: %u", sil, sil->minor_ver);
	DP("major_ver: %u", sil, sil->major_ver);
	DP("seconds: %u", sil, sil->seconds % 60);
	DP("seconds(full): 0x%x", sil, sil->seconds);
	DP("minutes: %u", sil, sil->minutes % 60);
	DP("minutes(full): 0x%x", sil, sil->minutes);
	DP("hour: %u", sil, sil->hour);
	DP("day: %u", sil, sil->day);
	DP("month: %u", sil, sil->month);
	DP("year: %u", sil, sil->year);
	DP("raid0_stride: %u", sil, sil->raid0_stride);
	DP("disk_number: %u", sil, sil->disk_number);
	DP("type: %u", sil, sil->type);
	DP("drives_per_striped_set: %d", sil, sil->drives_per_striped_set);
	DP("striped_set_number: %d", sil, sil->striped_set_number);
	DP("drives_per_mirrored_set: %d", sil, sil->drives_per_mirrored_set);
	DP("mirrored_set_number: %d", sil, sil->mirrored_set_number);
	DP("rebuild_ptr_low: %u", sil, sil->rebuild_ptr_low);
	DP("rebuild_ptr_high: %u", sil, sil->rebuild_ptr_high);
	DP("incarnation_no: %u", sil, sil->incarnation_no);
	DP("member_status: %u", sil, sil->member_status);
	DP("mirrored_set_state: %u", sil, sil->mirrored_set_state);
	DP("reported_device_location: %u", sil, sil->reported_device_location);
	DP("idechannel: %u", sil, sil->idechannel);
	DP("auto_rebuild: %u", sil, sil->auto_rebuild);
	DP("checksum1: %u", sil, sil->checksum1);
	DP("checksum2: %u", sil, sil->checksum2);
}
#endif

static struct dmraid_format sil_format = {
	.name	= "sil",
	.descr	= "(+) Silicon Image(tm) Medley(tm)",
	.format = FMT_RAID,
	.read	= sil_read,
	.write	= sil_write,
	.group	= sil_group,
	.check	= sil_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= sil_log,
#endif
	.type	= sil_type,
};

/* Register this format handler with the format core */
int register_sil(struct lib_context *lc)
{
	return register_format_handler(lc, &sil_format);
}

/* Set the RAID device contents up derived from the SIL ones */
static int _test_stripes(struct sil *sil)
{
	return sil->drives_per_striped_set > -1 &&
	       sil->disk_number < sil->drives_per_striped_set;
}

static int _test_mirrors(struct sil *sil)
{
	return sil->drives_per_mirrored_set > -1 &&
	       sil->disk_number < sil->drives_per_mirrored_set;
}

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct sil *sil, struct disk_info *di)
{
	unsigned int i;
	uint64_t array_sectors = (((uint64_t) sil->array_sectors_high) << 32) +
				 sil->array_sectors_low;
	struct meta_areas *ma;

	rd->di = di;
        rd->name = _name(lc, sil, sil->type == SIL_T_RAID10);
	rd->fmt = &sil_format;

	rd->offset = SIL_DATAOFFSET;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, AREAS)))
		return 0;

	for (ma = rd->meta_areas, i = 0; i < rd->areas; i++, ma++) {
		ma->offset = SIL_META_AREA(i) >> 9;
		ma->size = sizeof(*sil);
		ma->area = (void*) sil;
	}

	rd->status = _status(rd);

	switch (sil->type) {
	case SIL_T_RAID0:
		if (!_test_stripes(sil))
			return 0;

		rd->sectors = array_sectors / sil->drives_per_striped_set;
		break;

	case SIL_T_RAID1:
		if (!_test_mirrors(sil))
			return 0;

		rd->sectors = array_sectors;
		break;

	case SIL_T_RAID10:
		if (!_test_mirrors(sil) ||
		    !_test_stripes(sil))
			return 0;

		rd->sectors = array_sectors / sil->drives_per_mirrored_set;
		break;

	case SIL_T_SPARE:
		/* Cook them up... */
		rd->sectors = di->sectors - 3 * 512 - (di->sectors & 1) ? 1 : 2;
		break;

	default:
		log_err(lc, "%s: type: %u not supported", handler, sil->type);
		return 0;
	}

	return 1;
}
