/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include "internal.h"
#include "ondisk.h"

/*
 * Check that mandatory members of a metadata form handler are present.
 * 
 * We can only use log_err and log_print here, because
 * debug and verbose options are checked for later...
 */
static int check_format_handler(struct lib_context *lc,
				struct dmraid_format *fmt)
{
	unsigned int error = 0;
	const char *prefix = "missing metadata format handler ";

	if (!fmt)
		BUG(lc, 0, "%sNULL pointer", prefix);

	if (!fmt->name) {
		log_err(lc, "%sname", prefix);
		error++;
	}

	if (!fmt->descr) {
		log_err(lc, "%s: %sdescription", fmt->name, prefix);
		error++;
	}

	if (!fmt->read) {
		log_err(lc, "%s: %sread method", fmt->name, prefix);
		error++;
	}

	if (!fmt->group) {
		log_err(lc, "%s: %sgroup method", fmt->name, prefix);
		error++;
	}

	if (!fmt->check) {
		log_err(lc, "%s: %scheck method", fmt->name, prefix);
		error++;
	}

	if (!fmt->type) {
		log_err(lc, "%s: %stype method", fmt->name, prefix);
		error++;
	}

	if (fmt->format == FMT_RAID) {
		if (!fmt->write) {
			log_err(lc, "%s: %swrite method", fmt->name, prefix);
			error++;
		}

/* FIXME: uncomment after implementing event methods.
		if (!fmt->event) {
			log_err(lc, "%s: %sevent method", fmt->name, prefix);
			error++;
		}
*/

#ifdef	DMRAID_NATIVE_LOG
		if (!fmt->log) {
			log_err(lc, "%s: %slog method", fmt->name, prefix);
			error++;
		}
#endif
	}

	return error ? 0 : 1;
}

/*
 * Register a RAID device format handler.
 */
int register_format_handler(struct lib_context *lc, struct dmraid_format *fmt)
{
	struct format_list *fl;

	if (!check_format_handler(lc, fmt))
		return 0;

	if (!(fl = dbg_malloc(sizeof(*fl))))
		return 0;

	fl->fmt = fmt;
	list_add_tail(&fl->list, lc_list(lc, LC_FORMATS));

	return 1;
}

/* Return format name */
const char *get_format_name(struct lib_context *lc, void *fmt)
{
	return ((struct dmraid_format *) fmt)->name;
}

/*
 * Register all format handlers.
 *
 * I use an array because of the growing number...
 */
static struct register_fh {
	int (*func)(struct lib_context *lc);
} register_fh[] = {
#include "register.h"
	{ NULL },
};

void unregister_format_handlers(struct lib_context *lc)
{
	struct list_head *elem, *tmp;
	struct format_list *fl;

	list_for_each_safe(elem, tmp, lc_list(lc, LC_FORMATS)) {
		list_del(elem);
		dbg_free(list_entry(elem, typeof(*fl), list));
	}

	INIT_LIST_HEAD(lc_list(lc, LC_FORMATS));
}

int register_format_handlers(struct lib_context *lc)
{
	int ret = 1;
	struct register_fh *fh;

	for (fh = register_fh; fh->func; fh++) {
		if (!(ret = fh->func(lc)))
			goto bad;
	}

	return 1;

  bad:
	/* Clean up in case of error */
	log_err(lc, "failed registering format");
	unregister_format_handlers(lc);

	return ret;
}

/* Allocate private space in format handlers (eg, for on-disk metadata). */
void *alloc_private(struct lib_context *lc, const char *name, size_t size)
{
	void *ret;

	if ((ret = dbg_malloc(size)))
		memset(ret, 0, size);
	else
		log_err(lc, "failed allocating %s metadata", name);

	return ret;
}

/* Allocate metadata sector array in format handlers. */
void *alloc_meta_areas(struct lib_context *lc, struct dmraid_dev *rd,
		       const char *name, unsigned int n)
{
	void *ret;
	size_t size = n * sizeof(*rd->meta_areas);

	if ((ret = dbg_malloc(size))) {
		memset(ret, 0, size);
		rd->areas = n;
	} else
		log_err(lc, "failed allocating %s metadata areas array", name);

	return ret;
}

/* Simple metadata write function for format handlers */
static int _write_metadata(struct lib_context *lc, const char *handler,
			   struct dmraid_dev *rd, int idx, int erase)
{
	int ret = 0;
	void *p, *tmp;

	if (idx >= rd->areas)
		goto out;

	p = tmp = rd->meta_areas[idx].area;
	if (erase &&
	    !(p = alloc_private(lc, handler, rd->meta_areas[idx].size)))
		goto out;

	ret = write_file(lc, rd->di->path, (void *) p,
			 rd->meta_areas[idx].size,
			 rd->meta_areas[idx].offset << 9);

	log_dbg(lc, "writing metadata to %s, offset %" PRIu64 " sectors, size %zu "
		"sectors returned %d", rd->di->path, rd->meta_areas[idx].offset,
		rd->meta_areas[idx].size, ret); 
	if (p != tmp)
		dbg_free(p);

  out:
	return ret;
}

int write_metadata(struct lib_context *lc, const char *handler,
		   struct dmraid_dev *rd, int idx, int erase)
{
	unsigned int i;

	if (idx > -1)
		return _write_metadata(lc, handler, rd, idx, erase);

	for (i = 0; i < rd->areas; i++) {
		if (!_write_metadata(lc, handler, rd, i, erase))
			return 0;
	}

	return 1;
}

/*
 * Support function metadata format handlers.
 *
 * Check list of RAID sets.
 */
void check_rs(struct lib_context *lc, struct dmraid_set *rs,
	      int (*check)(struct lib_context *lc, struct dmraid_set *r))
{
	struct dmraid_set *r;

	if (list_empty(&rs->sets))
		return;

	list_for_each_entry(r, &rs->sets, list) {
		log_dbg(lc, "checking RAID set \"%s\"", r->name);
		check(lc, r);
	}
}

/* Check if format identifier is valid. */
int check_valid_format(struct lib_context *lc, char *name)
{
	struct format_list *fl;

	list_for_each_entry(fl, lc_list(lc, LC_FORMATS), list) {
		if (!strcmp(name, get_format_name(lc, fl->fmt)))
			return 1;
	}

	return 0;
}


/*
 * Update RAID set state based on operational subsets/devices.
 *
 * In case of a RAID set hierachy, check availability of subsets
 * and set superset to broken in case *all* subsets are broken.
 * If at least one is still available, set to inconsistent.
 *
 * In case of lowest level RAID stes, check consistence of devices
 * and make the above decision at the device level.
 */
static void _set_rs_status(struct lib_context *lc, struct dmraid_set *rs,
			   unsigned int i, unsigned int operational)
{
	if (operational == i)
		rs->status = s_operational;
	else if (operational > 0)
		rs->status = s_inconsistent;
	else
		rs->status = s_broken;

	log_dbg(lc, "set status of set \"%s\" to %u", rs->name, rs->status);
}

enum status set_rs_status(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int i = 0, operational = 0;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	/* Set status of subsets. */
	if (list_empty(&rs->devs)) {
		list_for_each_entry(r, &rs->sets, list) {
			/* Check subsets to set status of superset. */
			i++;
			if (r->status == s_operational ||
			    r->status == s_inconsistent)
				operational++;
	
		}
	} else {
		/* Check status of devices... */
		list_for_each_entry(rd, &rs->devs, list) {
			i++;
			if (rd->status == s_operational)
				operational++;
		}
	}
	
	_set_rs_status(lc, rs, i, operational);
	return rs->status;
}
