/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#ifndef	_FORMAT_H_
#define	_FORMAT_H_

/* Metadata format handler types. */
enum fmt_type {
	FMT_RAID,
	FMT_PARTITION,
};

#  ifdef FORMAT_HANDLER

#include <sys/types.h>
#include "datastruct/list.h"

struct disk_info;

/* Type definiton for metadata format handler events. */
enum event_type {
	et_undef = 0,
	et_io,		/* I/O error on a sector of a RAID device. */
	et_rd_add,	/* Add RAID device to set (eg, mirror replacement). */
	et_rd_remove,	/* Remove RAID device from set (eg, mirror failure). */
	/* ... */
};

/* Event: I/O error */
struct event_type_io {
	struct dmraid_set *rs;	/* RAID set of I/O error. */
	struct dmraid_dev *rd;	/* RAID device of I/O error. */
	uint64_t sector;	/* Sector of the I/O error. */
};

/* Event: RAID device add/remove */
struct event_type_rd {
	struct dmraid_set *rs;
	struct dmraid_dev *rd;
};

struct dmraid_format {
	const char *name;	/* Format name */
	const char *descr;	/* Format description */
	enum fmt_type format;	/* Format type (RAID, partition) */

	/*
	 * Read RAID metadata off a device and unify it.
	 */
	struct dmraid_dev* (*read)(struct lib_context *lc,
				   struct disk_info* di);

	/*
	 * Write RAID metadata to a device deunifying it
	 * or erase ondisk metadata if erase != 0.
	 */
	int (*write)(struct lib_context *lc, struct dmraid_dev* rd, int erase);

	/*
	 * Group a RAID device into a set.
	 */
	struct dmraid_set* (*group)(struct lib_context *lc,
				    struct dmraid_dev *rd);

	/*
	 * Check consistency of the RAID set metadata.
	 */
	int (*check)(struct lib_context *lc, struct dmraid_set *rs);

	/*
	 * Send an event to a metadata format handler (eg, I/O error).
	 *
	 * Returns 1 for event taken, RAID device write necessary.
	 * Returns 0 for error and/or write unnecessary.
	 */
	int (*event)(struct lib_context *lc, enum event_type type, void *arg);

	/*
	 * Display RAID disk metadata native.
	 */
	void (*log)(struct lib_context *lc, struct dmraid_dev *rd);

	/*
	 * Return unified layout type of RAID device.
	 */
	enum type (*type)(struct lib_context *lc, struct dmraid_dev *rd);
};

/* Chain of registered format handlers. */
struct format_list {
	struct list_head list;
	struct dmraid_format *fmt;
};

void *alloc_meta_areas(struct lib_context *lc, struct dmraid_dev *rd,
		       const char *name, unsigned int n);
void *alloc_private(struct lib_context *lc, const char *name, size_t size);
void check_rs(struct lib_context *lc, struct dmraid_set *rs,
	     int (*check)(struct lib_context *lc, struct dmraid_set *r));
int check_valid_format(struct lib_context *lc, char *fmt);
const char *get_format_name(struct lib_context *lc, void *fmt);
int register_format_handler(struct lib_context *lc, struct dmraid_format *fmt);
int register_format_handlers(struct lib_context *lc);
void unregister_format_handlers(struct lib_context *lc);
enum status set_rs_status(struct lib_context *lc, struct dmraid_set *rs);
int write_metadata(struct lib_context *lc, const char *handler,
		   struct dmraid_dev *rd, int meta_index, int erase);

#define	to_disk	to_cpu

#define	P(format, basevar, x, y...) \
	do { \
		log_print(lc, "0x%03x " format, \
			  (unsigned long) &x - (unsigned long) basevar, y); \
	} while(0)

#define	P2(format, basevar, i, x) do { P(format, basevar, x, i, x); } while(0)
#define	DP(format, basevar, x) do { P(format, basevar, x, x); } while(0)

#  endif /* ifdef FORMAT_HANDLER */

#endif
