/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * DOS partition format handler.
 *
 * dos_read() and dos_group influenced by libparted.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "dos.h"
#include <datastruct/byteorder.h>

static const char *handler = "dos";

/* Make up RAID device name. */
static size_t __name(struct dmraid_dev *rd, unsigned int partition,
		     char *str, size_t len, unsigned char type)
{
	char *p, part[5];

	p = strrchr(rd->di->path, '/');
	p = p ? p + 1 : rd->di->path;
	snprintf(part, 5, "_p%u", partition);
	return snprintf(str, len, "%s%s", p, type ? part : "") + 1;
}

static char *_name(struct lib_context *lc, struct dmraid_dev *rd,
		   unsigned int partition, unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(rd, partition, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1)))
		__name(rd, partition, ret, len, type);
	else
		log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/* Mapping of DOS type to generic type. */
static enum type dos_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return t_partition;
}

/*
 * Read a DOS partiton.
 */
/* Endianess conversion. */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct dos *dos)
{
	struct dos_partition *part = dos->partitions;

	for (part = dos->partitions; part < &dos->partitions[4]; part++) {
		part->start  = le32_to_cpu(part->start);
		part->length = le32_to_cpu(part->length);
	}

	dos->magic = le16_to_cpu(dos->magic);
}
#endif

/* Allocate a DOS partition sector struct and read the data. */
static int _alloc_and_read(struct lib_context *lc, struct dos **dos,
			   char *path, uint64_t sector)
{
	struct dos *d;

	if (!(d = alloc_private(lc, handler, sizeof(*d))))
		return 0;

	if (!read_file(lc, path, (void *) d, sizeof(*d), sector << 9)) {
		dbg_free(d);
		return 0;
	}

	to_cpu(d);
	*dos = d;

	return 1;
}

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct dos *dos, struct disk_info *di);
static struct dmraid_dev *dos_read(struct lib_context *lc, struct disk_info *di)
{
	struct dos *dos;
	struct dos_partition *p;
	struct dmraid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata */
	if (!_alloc_and_read(lc, &dos, di->path, DOS_CONFIGOFFSET >> 9))
		goto out;

	/* Check magic. */
	if (dos->magic != DOS_MAGIC)
		goto bad;

	for (p = dos->partitions; p < &dos->partitions[4]; p++) {
		if (p->type == PARTITION_GPT)
			goto bad;
	}

	if (lc_opt(lc, LC_DUMP))
		log_print(lc, "%s: filing metadata not supported (use fdisk "
			  "and friends)", handler);

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_dev(lc)))
		goto bad;

	/*
	 * Setup generic RAID device structure members using
	 * ondisk inforation.
	 */
	if (setup_rd(lc, rd, dos, di))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(dos);

   out:
	return rd;
}

/* Support functions for dos_group to read the partition table(s). */
static int part_is_extended (struct dos_partition *part)
{
	switch (part->type) {
	case PARTITION_EXT:
	case PARTITION_EXT_LBA:
	case PARTITION_LINUX_EXT:
		return 1;
	}

	return 0;
}

/* Get a device offset relative to a partition. */
static uint64_t linear_start(const struct dos_partition *raw_part,
			     uint64_t offset)
{
	return (uint64_t) raw_part->start + offset;
}

/* RAID set allocation support function. */
static struct dmraid_set *_alloc_set(struct lib_context *lc,
				     struct dmraid_dev *rd)
{
	struct dmraid_set *rs;

	if ((rs = find_set(lc, rd->name, 0)))
		LOG_ERR(lc, NULL, "%s: RAID set %s already exists",
			handler, rs->name);
	else {
		if (!(rs = alloc_set(lc)))
			LOG_ERR(lc, NULL, "%s: Failed allocating RAID set",
				handler);

		rs->type = dos_type(lc, rd);
		rs->status = s_operational;
	}

	return rs;
}

/*
 * Allocate a DOS RAID device and a set, set
 * the device up and add it to the set.
 */
static int _create_rs_and_rd(struct lib_context *lc, struct dmraid_dev *rd,
			     struct dos_partition *raw_part, uint64_t sector,
			     unsigned int part)
{
	struct dmraid_dev *r;
	struct dmraid_set *rs;

	if (!(r = alloc_dev(lc)))
		return 0;

	if (!(rs = _alloc_set(lc, r)))
		goto free_dev;

	if (!(r->di = alloc_disk_info(rd->di->path)))
		goto free_set;

        r->di->serial = NULL;
	r->fmt = rd->fmt;
	r->status = rd->status;

        r->name = _name(lc, rd, part, 1);
	if ((uint64_t) raw_part->start > sector)
		sector = 0;

	r->offset = linear_start(raw_part, sector);
	r->sectors = (uint64_t) raw_part->length;

        if (!(rs->name = dbg_strdup(r->name)))
		goto free_di;

	list_add_tail(&r->list, &rs->devs);
	list_add_tail(&rs->list, lc_list(lc, LC_RAID_SETS));

	return 1;

  free_di:
	free_disk_info(lc, r->di);
  free_set:
	free_set(lc, rs);
  free_dev:
	free_dev(lc, r);
	return 0;
}

/*
 * RAID set grouping.
 *
 * In this case it is not really about grouping, more about providing
 * the propper input to the activation layer by creating a RAID set per
 * logical partition and a RAID dev hanging off it mapping the partition
 * linearly.
 *
 * Partition code inspired by libparted and squeezed for this purpose.
 */
static char out_of_order;
static uint64_t lo2;
static int is_partition(struct dos_partition *p, uint64_t start_sector)
{
	return p->type != PARTITION_EMPTY && p->length &&
	       (start_sector ? out_of_order ||
			       ((uint64_t) p->start < start_sector) : 1);
}

/* FIXME: skip combined/deleted/recreated partitions. */
static uint64_t next_partition(uint64_t start_sector,
			       struct dos_partition *p1,
			       struct dos_partition *p2)
{
	uint64_t o1, o2, o3, ret;

	o1 = linear_start(p1, start_sector + (uint64_t) p1->length);
	o2 = linear_start(p2, (uint64_t) p2->length);
	o3 = start_sector + (uint64_t) p2->length;

	if (out_of_order && start_sector == lo2)
		ret = start_sector;
	else if ((uint64_t) p1->start > start_sector)
		ret = o3;
	else if (o2 < start_sector)
		ret = o2;
	else if (o1 == o3)
		ret = o2;
	else
		ret = o1;

	lo2 = o2;
	if (ret < start_sector)
		out_of_order = 1;

	return ret;
}

static int group_rd_extended(struct lib_context *lc, struct dmraid_dev *rd,
			     uint64_t start_sector, unsigned int part)
{
	int ret = 0;
	uint64_t new_start_sector;
	struct dos *dos;
	struct dos_partition *p1, *p2;

	/* Allocate and read an extended partiton table. */
	if (!_alloc_and_read(lc, &dos, rd->di->path, start_sector))
		return 0;

	/* Weird: empty extended partitions are filled with 0xF6 by PM */
	if (dos->magic == PARTITION_MAGIC_MAGIC)
		goto out;

	p1 = &dos->partitions[0];
	p2 = &dos->partitions[1];

	if (is_partition(p1, start_sector)) {
		if (!_create_rs_and_rd(lc, rd, p1, start_sector, part++))
			goto out;
	}

	/* Recurse into the partition chain. */
	if ((new_start_sector = next_partition(start_sector, p1, p2))) {
		if (new_start_sector != start_sector &&
	    	    !group_rd_extended(lc, rd, new_start_sector, part))
			goto out;
	}

	ret = 1;

  out:
	dbg_free(dos);
	return ret;
}

/* Handle primary partitions. */
static int group_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    uint64_t start_sector)
{
	unsigned int i;
	uint64_t part_start, extended_part_start = 0;
	struct dos *dos = rd->meta_areas[0].area;
	struct dos_partition *raw_part;

	out_of_order = lo2 = 0;

	/*
	 * Walk 4 array entries of a primary partition table.
	 *
	 * FIXME: check integrity of partition's C/H/S values.
	 */
	for (i = 0; i < 4; i++) {
		raw_part = &dos->partitions[i];
		if (!is_partition(raw_part, start_sector))
			continue;

		part_start = linear_start(raw_part, start_sector);

		/* Avoid infinite recursion. */
		if (part_start == start_sector || part_start > rd->sectors)
			continue;

		if (part_is_extended(raw_part))
			extended_part_start = part_start;
		else if (!_create_rs_and_rd(lc, rd, raw_part,
					    start_sector, i + 1))
			return 0;
	}

	return extended_part_start ?
	       group_rd_extended(lc, rd, extended_part_start, 5) : 1;
}

/* Add a DOS RAID device to a set */
static struct dmraid_set *dos_group(struct lib_context *lc,
				    struct dmraid_dev *rd)
{
	/*
	 * Once we get here, a DOS boot sector with a partition table
	 * has been discovered by dos_read.
	 *
	 * We need to run through the (nested) partition tables and create
	 * a RAID set and a linear RAID device hanging off it for every primary
	 * and logical partition, so that the activate code is happy.
	 */
	return group_rd(lc, rd, 0) ? (struct dmraid_set *) 1 : NULL;
}

/*
 * Check integrity of a DOS RAID set.
 */
static int dos_check(struct lib_context *lc, struct dmraid_set *rs)
{
	return 1; /* Nice, eh ? */
}

static struct dmraid_format dos_format = {
	.name	= "dos",
	.descr	= "(+) DOS partitions on SW RAIDs",
	.format = FMT_PARTITION,
	.read	= dos_read,
	.write	= NULL, /* Not supported */
	.group	= dos_group,
	.check	= dos_check,
	.type	= dos_type,
};

/* Register this format handler with the format core */
int register_dos(struct lib_context *lc)
{
	return register_format_handler(lc, &dos_format);
}

/*
 * Set the RAID device contents up derived from the DOS ones
 *
 * For a DOS partition we essentially just save the partition table sector
 * and let dos_group do the rest...
 */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct dos *dos, struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(lc, rd, 0, 0);
	rd->fmt = &dos_format;
	rd->offset = DOS_DATAOFFSET;
	rd->sectors = di->sectors;

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = DOS_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*dos);
	rd->meta_areas[0].area = (void *) dos;

        rd->status = s_operational; /* Always :) */

	return 1;
}
