/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#include "internal.h"

/*
 * Levels definitions.
 *
 * Must be in the same order as in the type enum !!!
 */
const char *ascii_types[] = {
	NULL,
	"GROUP",
	"partition",
	"spare",
	"linear",
	"stripe",
	"mirror",
	/* RAID types below not supported (yet) */
	"raid3",
	"raid4",
	"raid5",
};

/*
 * Stacked types definitions.
 *
 * get_stacked_type() below depending on order!
 */
const char *stacked_ascii_types[] = {
	"raid10",
	"raid30",
	"raid40",
	"raid50",
	"raid01",
	"raid03",
	"raid04",
	"raid05",
};

/*
 * Status definitions.
 *
 * Must be in the same order as in the status enum !!!
 */
const char *status[] = {
	NULL,
	"setup",
	"operational",
	"broken",
	"nosync",
	"inconsistent",
};

/* Fetch the respective ASCII string off the types array. */
const char *get_type(struct lib_context *lc, enum type l)
{
	return l >= ARRAY_SIZE(ascii_types) ? NULL : ascii_types[l];
}

/* Check, if a RAID set is stacked. */
static int stacked_set(struct dmraid_set *rs)
{
	return list_empty(&rs->sets) ? 0 : 1;
}

/* Return the RAID type of a stacked RAID set (eg, raid10). */
static const char *get_stacked_type(void *v)
{
	unsigned int i;
	struct dmraid_set *rs = v;
	enum type type = rs->type;

	if (type == t_raid0) {
		if (list_empty(&rs->sets))
			return NULL;

		i =  ARRAY_SIZE(stacked_ascii_types) / 2 - 1;
		type = list_entry(rs->sets.next, struct dmraid_set, list)->type;
	} else
		i = 0;

	switch (type) {
	case t_raid5:
		i++;
	case t_raid4:
		i++;
	case t_raid3:
		i++;
	case t_raid1:
		break;

	default:
		return NULL;
	}

	return stacked_ascii_types[i];
}

/* Return the ASCII type for a RAID set. */
const char *get_set_type(struct lib_context *lc, void *v)
{
	struct dmraid_set *rs = v;

	if (rs->type != t_group && stacked_set(rs))
		return get_stacked_type(rs);

	return get_type(lc, rs->type);
}

/* Fetch the respective ASCII string off the types array. */
const char *get_status(struct lib_context *lc, void *rs)
{
	enum status s = ((struct dmraid_set *) rs)->status;

	return s >= ARRAY_SIZE(status) ? NULL : status[s];
}

/*
 * Calculate the size of the set by recursively
 * summing the size of the devices in the subsets.
 *
 * Pay attention to RAID > 0 types.
 */
uint64_t calc_total_sectors(struct lib_context *lc, struct dmraid_set *rs)
{
        uint64_t sectors = 0;
        struct dmraid_set *r;
        struct dmraid_dev *rd;

	if (rs->type != t_group) {
		list_for_each_entry(r, &rs->sets, list) {
	                sectors += calc_total_sectors(lc, r);
			if (rs->type == t_raid1)
				break;
		}
	}

	list_for_each_entry(rd, &rs->devs, list) {
                sectors += rd->sectors;
		if (rs->type == t_raid1)
			break;
	}

	if (rs->type == t_raid3 || rs->type == t_raid5) {
		sectors -= list_empty(&rs->sets) ?
			   list_entry(rs->devs.next, typeof(*rd), list)->sectors :
			   calc_total_sectors(lc, list_entry(rs->sets.next, typeof(*rs), list));
	} 

        return sectors;
}

/* Count devices in a set recursively. */
unsigned int count_devs(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int ret = 0;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, list) {
		if (rs->type != t_group)
			ret += count_devs(lc, r);
	}

	list_for_each_entry(rd, &rs->devs, list)
		ret++;

	return ret;
}

/*
 * Create list of unique memory pointers of a RAID device and free them.
 *
 * This prevents me from having a destructor method in the metadata
 * format handlers so far. If life becomes more complex, I might need one
 * though...
 */
static void _free_pointers(struct lib_context *lc, struct dmraid_dev *rd)
{
	int area, i, idx = 0;
	size_t size;
	void **p;

	if (!(area = (rd->private ? 1 : 0) + rd->areas))
		return;

	size = (area + 1) * sizeof(*p);
	if ((p = dbg_malloc(size)))
		memset(p, 0, size);
	else
		LOG_ERR(lc, , "allocating pointer array");

	if (rd->private)
		p[idx++] = rd->private;

	if (rd->areas) {
		for (area = 0; area < rd->areas; area++) {
			for (i = 0; i < idx; i++) {
				if (p[i] == rd->meta_areas[area].area)
					break;
			}
	
			if (i == idx)
				p[idx++] = rd->meta_areas[area].area;
		}

		dbg_free(rd->meta_areas);
	}

	for (i = 0; i < idx; i++)
		dbg_free(p[i]);
	
	dbg_free(p);
}

/* Allocate/Free RAID device (member of a RAID set). */
struct dmraid_dev *alloc_dev(struct lib_context *lc)
{
	struct dmraid_dev *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret));
		INIT_LIST_HEAD(&ret->list);
	} else
		log_err(lc, "failed allocating RAID device");

	return ret;
}

void free_dev(struct lib_context *lc, struct dmraid_dev *rd)
{
	/*
	 * Create list of memory pointers allocated by the metadata
	 * format handler and free them.
	 */
	_free_pointers(lc, rd);
	dbg_free(rd->name);
	dbg_free(rd);
}

/* Allocate/Free RAID set. */
struct dmraid_set *alloc_set(struct lib_context *lc)
{
	struct dmraid_set *ret;

	if ((ret = dbg_malloc(sizeof(*ret)))) {
		memset(ret, 0, sizeof(*ret)); /* Clear needed! */
		INIT_LIST_HEAD(&ret->list);
		INIT_LIST_HEAD(&ret->sets);
		INIT_LIST_HEAD(&ret->devs);
	} else
		log_err(lc, "failed allocating RAID set");

	return ret;
}

/* Remove a set or all sets (in case rs = NULL) recursively. */
/* FIXME: remove partition code in favour of kpartx ? */
static void _free_set(struct lib_context *lc, struct dmraid_set *rs)
{
	struct list_head *elem, *tmp;
	struct dmraid_dev *rd;

	if (rs)
		log_dbg(lc, "freeing devices of RAID set \"%s\"", rs->name);

	list_for_each_safe(elem, tmp,
			   rs ? &rs->sets : lc_list(lc, LC_RAID_SETS))
		free_set(lc, list_entry(elem, struct dmraid_set, list));

	if (rs) {
		list_for_each_safe(elem, tmp, &rs->devs) {
			list_del(elem);
			rd = list_entry(elem, struct dmraid_dev, list);
	
			/* Prevent free_disk_info() from freeing again. */
			rd->di->rd = NULL;
	
			log_dbg(lc, "freeing device \"%s\", path \"%s\"",
				rd->name, rd->di->path);

			if (partitioned_set(lc, rs))
				free_disk_info(lc, rd->di);

			free_dev(lc, rd);
		}
	
		list_del(&rs->list);
		dbg_free(rs->name);
		dbg_free(rs);
	} else if (!list_empty(lc_list(lc, LC_RAID_SETS)))
		log_fatal(lc, "lib context RAID set list not empty");
}

void free_set(struct lib_context *lc, struct dmraid_set *rs)
{
	/* Handle deletion of virgin RAID set structures. */
	if (rs && !rs->name) {
		if (!list_empty(&rs->sets) ||
		    !list_empty(&rs->devs))
			BUG(lc, , "no virgin RAID set!!!");

		dbg_free(rs);
	} else
		_free_set(lc, rs);
}

/* Return != 0 in case of a partitioned RAID set type. */
int partitioned_set(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set*) rs)->type == t_partition;
}

/* Return != 0 in case of a partitioned base RAID set. */
int base_partitioned_set(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set*) rs)->has_partitions;
}

/* Return RAID set name. */
const char *get_set_name(struct lib_context *lc, void *rs)
{
	return ((struct dmraid_set *) rs)->name;
}

/* Find RAID set by name (recursive if decend != 0). */
static struct dmraid_set *_find_set(struct lib_context *lc,
				    struct list_head *list,
				    const char *name, int decend) {
	struct dmraid_set *rs;

	if (list_empty(list) || !name)
		return NULL;

	log_dbg(lc, "_find_set: searching %s", name);

	list_for_each_entry(rs, list, list) {
		if (!strcmp(rs->name, name))
			goto out;
	}

	if (decend) {
		list_for_each_entry(rs, list, list) {
			struct dmraid_set *r;

			if ((r = _find_set(lc, &rs->sets, name, decend))) {
				rs = r;
				goto out;
			}
		}
	}

	return NULL;

  out:
	log_dbg(lc, "_find_set: found %s", name);
	return rs;
}

struct dmraid_set *find_set(struct lib_context *lc,
			    const char *name, int decend)
{
	return _find_set(lc, lc_list(lc, LC_RAID_SETS), name, decend);
}

/* Return # of raid sets build */
static unsigned int _count_sets(struct lib_context *lc)
{
	int ret = 0;
	struct list_head *elem;

	list_for_each(elem, lc_list(lc, LC_RAID_SETS))
		ret++;

	return ret;
}

/*
 * Count devices found
 */
static unsigned int _count_devices(struct lib_context *lc, enum dev_type type)
{
	unsigned int ret = 0;
	struct disk_info *di;

	list_for_each_entry(di, lc_list(lc, LC_DISK_INFOS), list) {
		if (DEVICE == type)
			ret++;
		else if (((RAID|NATIVE) & type) && di->rd)
			ret++;
	}

	return ret;
}

unsigned int count_devices(struct lib_context *lc, enum dev_type type)
{
	return type == SET ? _count_sets(lc) : _count_devices(lc, type);
}

/*
 * Read RAID metadata off a device by trying
 * all registered format handlers in turn.
 */
static int _want_format(struct dmraid_format *fmt, const char *format,
			enum fmt_type type)
{
	if (fmt->format != type)
		return 0;
	else if (format && strcmp(format, fmt->name))
		return 0;

	return 1;
}

static struct dmraid_dev *_dmraid_read(struct lib_context *lc,
				       struct disk_info *di,
				       struct dmraid_format *fmt)
{
	struct dmraid_dev *rd;

	log_info(lc, "%s: %-6s discovering", di->path, fmt->name);

	if ((rd = fmt->read(lc, di))) {
		log_info(lc, "%s: %s metadata discovered",
			 di->path, fmt->name);
		rd->fmt = fmt;
	}

	return rd;
}

static struct dmraid_dev *dmraid_read(struct lib_context *lc,
				      struct disk_info *di, char const *format,
				      enum fmt_type type)
{
	struct format_list *fl;
	struct dmraid_dev *rd = NULL, *rd_tmp = NULL;

	list_for_each_entry(fl, lc_list(lc, LC_FORMATS), list) {
		if (_want_format(fl->fmt, format, type) &&
		    (rd_tmp = _dmraid_read(lc, di, fl->fmt))) {
			if (rd) {
				log_warn(lc, "%s: \"%s\" and \"%s\" formats "
					 "discovered (using %s)!",
					 di->path, rd_tmp->fmt->name,
					 rd->fmt->name, rd->fmt->name);
				free_dev(lc, rd_tmp);
			} else
				rd = rd_tmp;
		}
	}

	if (rd_tmp && rd_tmp != rd) {
		free_dev(lc, rd);
		rd = NULL;
	}

	return rd;
}

/*
 * Write RAID metadata to a device.
 */
static int dmraid_write(struct lib_context *lc,
			struct dmraid_dev *rd, int erase)
{
	int ret = 0;
	struct dmraid_format *fmt = rd->fmt;

	if (fmt->write) {
		log_notice(lc, "%s %s",
			   erase ? "Erasing metadata on" :
				   "Writing metadata to %s",
			   rd->di->path);
		ret = fmt->write(lc, rd, erase);
	} else
		log_err(lc, "format \"%s\" doesn't support writing metadata",
			fmt->name);

	return ret;
}

/*
 * Group RAID device into a RAID set.
 */
static struct dmraid_set *dmraid_group(struct lib_context *lc,
				       struct dmraid_dev *rd)
{
	return rd->fmt->group(lc, rd);
}

/* Check that device names are members of the devices list. */
static int _device_hit(struct disk_info *di, char **devices)
{
	char **dev;

	if (!devices || !*devices)
		return 1;

	for (dev = devices; *dev; dev++) {
		if (!strcmp(*dev, di->path))
			return 1;
	}

	return 0;
}

/* Discover RAID devices. */
void discover_raid_devices(struct lib_context *lc, char **fmt_names, char **devices)
{
	unsigned int i;
	struct disk_info *di;
	struct dmraid_dev *rd;

	list_for_each_entry(di, lc_list(lc, LC_DISK_INFOS), list) {
		if (_device_hit(di, devices)) {
			i = 0;
			do {
				if ((rd = dmraid_read(lc, di, fmt_names ? fmt_names[i] : NULL, FMT_RAID)))
					di->rd = rd;
			} while (fmt_names && fmt_names[++i]);
		}
	}
}

/*
 * Discover partitions on RAID sets.
 *
 * FIXME: remove partition code in favour of kpartx ?
 */
static void _discover_partitions(struct lib_context *lc,
				 struct list_head *rs_list)
{
	char *path;
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs, *r;

	list_for_each_entry(rs, rs_list, list) {
		/*
		 * t_group type RAID sets are never active!
		 * (They are containers for subsets to activate)
		 *
		 * Recurse into them.
		 */
		if (rs->type == t_group) {
			_discover_partitions(lc, &rs->sets);
			return;
		}

		/*
		 * Skip all "container" sets, which are not active.
		 */
		if (base_partitioned_set(lc, rs) ||
		    partitioned_set(lc, rs) ||
		    !dm_status(lc, rs))
			continue;

		log_dbg(lc, "discovering partitions on \"%s\"", rs->name);
		if (!(path = mkdm_path(lc, rs->name)))
			return;

		/* Allocate a temporary disk info struct for dmraid_read(). */
		di = alloc_disk_info(path);
		dbg_free(path);
		if (!di)
			return;

		di->sectors = calc_total_sectors(lc, rs);
		if (!(rd = dmraid_read(lc, di, NULL, FMT_PARTITION))) {
			free_disk_info(lc, di);
			continue;
		}

		/*
		 * WARNING: partition group function returns
		 * a dummy pointer because of the creation of multiple
		 * RAID sets (one per partition) it does.
		 *
		 * We don't want to access that 'pointer'!
		 */
		if ((r = dmraid_group(lc, rd))) {
			log_notice(lc, "created partitioned RAID set(s) for %s",
				   di->path);
			rs->has_partitions = 1;
		} else
			log_err(lc, "failed adding %s to RAID set", path);

		/*
		 * Free the RD. We don't need it any more, because we
		 * don't support writing partition tables.
		 */
		free_disk_info(lc, rd->di);
		free_dev(lc, rd);
	}
}

void discover_partitions(struct lib_context *lc)
{
	_discover_partitions(lc, lc_list(lc, LC_RAID_SETS));
}

/* Get format handler of RAID set. */
static struct dmraid_format *_get_fmt(struct dmraid_set *rs)
{
	struct dmraid_dev *rd;

	/* Decend RAID set hierarchy. */
	while (!list_empty(&rs->sets))
		rs = list_entry(rs->sets.next, typeof(*rs), list);

	rd = list_entry(rs->devs.next, typeof(*rd), list);
	return rd ? rd->fmt : NULL;
}

/* Check metadata consistency of raid sets. */
static void check_sets(struct lib_context *lc)
{
	struct list_head *elem, *tmp;
	struct dmraid_set *rs;
	struct dmraid_format *fmt;

	list_for_each_safe(elem, tmp, lc_list(lc, LC_RAID_SETS)) {
		rs = list_entry(elem, typeof(*rs), list);
		if (!(fmt = _get_fmt(rs)))
			continue;

		if (!fmt->check) {
			log_fatal(lc, "format \"%s\" lags a check method!",
				fmt->name);
			free_set(lc, rs);
			continue;
		}

		if (!fmt->check(lc, rs)) {
			log_err(lc, "RAID set \"%s\" is inconsistent", rs->name);
			/*
			 * FIXME: check needed if degraded activation
			 *        is sensible.
			 */
			if (rs->type == t_raid1)
				log_err(lc, "check of mirrored set \"%s\" failed, "
					"keeping it", rs->name);
			else {
				log_err(lc, "check of \"%s\" failed, "
					"removing RAID set", rs->name);
				free_set(lc, rs);
			}
		}
	}

	return;
}

/*
 * Group RAID set(s)
 *
 *	name = NULL  : build all sets
 *	name = String: build just the one set
 */
static void want_set(struct lib_context *lc, struct dmraid_set *rs, char *name)
{
	size_t len;

	if (!name)
		return;

	len = min(strlen(rs->name), strlen(name));
	if (strncmp(rs->name, name, len)) {
		log_info(lc, "dropping unwanted RAID set \"%s\"", rs->name);
		free_set(lc, rs);
	}
}

int group_set(struct lib_context *lc, char *name)
{
	struct disk_info *di;
	struct dmraid_dev *rd;
	struct dmraid_set *rs;

	if (name && find_set(lc, name, 0))
		LOG_ERR(lc, 0, "RAID set %s already exists\n", name);

	list_for_each_entry(di, lc_list(lc, LC_DISK_INFOS), list) {
		if (!(rd = di->rd))
			continue;
	
		if ((rs = dmraid_group(lc, rd))) {
			log_info(lc, "added %s to RAID set \"%s\"",
				 rd->di->path, rs->name);
			want_set(lc, rs, name);
		} else {
			if (rd->sectors)
				log_err(lc, "failed adding %s to RAID "
					"set \"%s\"",
					rd->di->path, rd->name);
			else
				log_print(lc, "skipping spare %s",
					  rd->di->path);

			/* Need to find the failed set and remove it. */
			if ((rs = find_set(lc, rd->name, 0))) {
				log_err(lc, "removing RAID set \"%s\"",
					rs->name);
				free_set(lc, rs);
			}
		}
	}

	/* Check sanity of grouped RAID sets. */
	check_sets(lc);

	return 1;
}

/* Process function on RAID set(s) */
static void process_set(int (*func)(struct lib_context *l, void *v, int a),
			struct lib_context *lc, void *rs, int arg)
{
	if (!partitioned_set(lc, rs))
		func(lc, rs, arg);
}

/* FIXME: remove partition code in favour of kpartx ? */
static void
process_partitioned_set(int (*func)(struct lib_context *l, void *v, int a),
			struct lib_context *lc, void *rs, int arg)
{
	if (partitioned_set(lc, rs) && !base_partitioned_set(lc, rs))
		func(lc, rs, arg);
}

void process_sets(int (*func)(struct lib_context *l, void *v, int a),
		  struct lib_context *lc, int arg, int partitions)
{
	struct dmraid_set *rs;
	void (*p)(int (*f)(struct lib_context *l, void *v, int a),
		  struct lib_context *l, void *v, int a) =
		partitions ? process_partitioned_set : process_set;

	list_for_each_entry(rs, lc_list(lc, LC_RAID_SETS), list)
		p(func, lc, rs, arg);
}

/* Write RAID set metadata to devices. */
int write_set(struct lib_context *lc, void *v)
{
	int ret = 0;
	struct dmraid_set *r, *rs = v;
	struct dmraid_dev *rd;

	/* Decend hierarchy */
	list_for_each_entry(r, &rs->sets, list) {
		/*
		 * FIXME: does it make sense to try the rest of the subset
		 *	  in case we fail writing one ?
		 */
		if (!write_set(lc, (void *) r))
			log_err(lc, "failed writing RAID subset \"%s\", continuing",
				r->name);
	}

	/* Write metadata to devices */
	list_for_each_entry(rd, &rs->devs, list) {
		if (!(ret = dmraid_write(lc, rd, 0)))
			break;
	}

	return ret;
}

/* Erase ondisk metadata. */
int erase_metadata(struct lib_context *lc)
{
	struct disk_info *di;

	list_for_each_entry(di, lc_list(lc, LC_DISK_INFOS), list) {
		if (di->rd &&
		    yes_no_prompt("Do you really want to erase \"%s\" ondisk "
				  "metadata on %s",
				  di->rd->fmt->name, di->path) &&
		    !dmraid_write(lc, di->rd, 1))
			log_err(lc, "erasing ondisk metadata on %s", di->path);
	}

	return 1;
}

/*
 * Support function for format handlers:
 *
 * Return neutralized RAID type for given mapping array (linear, raid0, ...)
 */
enum type rd_type(struct types *t, unsigned int type)
{
	for (; t->unified_type != t_undef; t++) {
		if (type == t->type)
			return t->unified_type;
	}

	return t_undef;
}

/*
 * Support function for metadata format handlers.
 *
 * Sort a device into a device list by optionally
 * using a metadata format handler helper function.
 */
void list_add_dev_sorted(struct lib_context *lc, struct dmraid_set *rs,
			 struct dmraid_dev *new,
			 int (*f)(struct dmraid_dev *rd,
				  struct dmraid_dev *new))
{
	struct dmraid_dev *rd;

	list_for_each_entry(rd, &rs->devs, list) {
		/*
		 * Add in at the beginning or where the metadata
		 * format handler function tells us to.
		 */
		if (!f || f(rd, new)) {
			list_add(&new->list, &rd->list);
			return;
		}
	}

	list_add_tail(&new->list, &rs->devs);
}

/*
 * Support function for format handlers:
 *
 * File RAID metadata and offset on device for analysis.
 */
static size_t __name(char *str, size_t len, char *name,
		     const char *suffix, const char *handler)
{
	char *s;
	
	s = strrchr(name, '/');
	s = s ? s + 1 : name;
	return snprintf(str, len, "%s_%s.%s", s, handler, suffix) + 1;
}

static char *_name(struct lib_context *lc, char *name,
		   const char *suffix, const char *handler)
{
	size_t len = __name(NULL, 0, name, suffix, handler) + 1;
	char *n;

	if ((n = dbg_malloc(len)))
		__name(n, len, name, suffix, handler);
	else
		log_err(lc, "allocation metadata file memory");

	return n;
}

static int _file_data(struct lib_context *lc, const char *handler,
		      char *path, void *data, size_t size)
{
	int ret;
	char *name;

	if (!(name = _name(lc, path, "dat", handler)))
		return 0;
	
	log_notice(lc, "writing metadata file %s", name);
	ret = write_file(lc, name, data, size, 0);
	dbg_free(name);

	return ret;
}

static int _file_offset(struct lib_context *lc, const char *handler,
			char *path, uint64_t offset)
{
	int ret;
	size_t size;
	char *name, s_offset[32];
	
	if (!(name = _name(lc, path, "offset", handler)))
		return 0;

	log_notice(lc, "writing metadata offset file %s", name);
	size = snprintf(s_offset, 32, "%" PRIu64, offset);
	ret = write_file(lc, name, (void *) s_offset, size, 0);
	dbg_free(name);

	return ret;
}

void file_metadata(struct lib_context *lc, const char *handler,
		   char *path, void *data, size_t size, uint64_t offset)
{
	if (lc_opt(lc, LC_DUMP)) {
		if (!_file_data(lc, handler, path, data, size))
			return;

		_file_offset(lc, handler, path, offset);
	}
}
