/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

#ifndef	_METADATA_H_
#define	_METADATA_H_

#include "datastruct/list.h"
#include <stdint.h>

/*
 * Unified RAID set types.
 *
 * Keep the order, some functions rely on it.
 */
enum type {
	t_undef	= 0,
	t_group,
	t_partition,
	t_spare,
	t_linear,
	t_raid0,
	t_raid1,
	/* Higher RAID types below not supported (yet). */
	t_raid3,
	t_raid4,
	t_raid5,
};

/*
 * Mapping struct for RAID type unification.
 *
 * Format handler allocates an array and inserts mappings
 * from format specific types to the unified ones above.
 */
struct types {
	unsigned int type; /* Must be long enough for vendor definition. */
	enum type    unified_type;
};

/* RAID disk status. */
enum status {
	s_undef = 0,
	s_setup,	/* During setup transition. */
	s_operational,
	s_broken,
	s_nosync,
	s_inconsistent,
	s_spare,
};

/* Metadata areas and size stored on a RAID device. */
struct meta_areas {
	uint64_t	offset;	/* on disk metadata offset in sectors. */
	size_t		size;	/* on disk metadata size in bytes. */
	void		*area;	/* pointer to format specific metadata. */
};

/*
 * Abstracted RAID device.
 *
 * A RAID device is a member of a RAID set and can only
 * exist at the lowest level of a RAID set stack (eg, for RAID10).
 */
struct dmraid_dev {
	struct list_head list;		/* chain of devices hanging off a set */
	struct disk_info  *di;		/* backpointer to disk_info */

	char		*name;		/* Name of set this device belongs to */

	struct dmraid_format *fmt;	/* Format handler for this device */

	enum status status;		/* Status of device */

	uint64_t	offset;		/* Data offset on device */
	uint64_t	sectors;	/* Length of the segment to map */

	unsigned int	areas;		/* # of metadata areas on the device */
	struct meta_areas *meta_areas;	/* Dynamic array of metadata areas */

	void		*private;	/* For format handler use (eg, to
					   keep references between calls) */
};

/*
 * Abstracted (hierarchical) RAID set.
 *
 * Can be used to form a tree of subsets with arbitrary depths.
 * Defines RAID attributes for the set as a whole (eg: RAID0, Status).
 */
struct dmraid_set {
	struct list_head list;		/* Chain of independent sets. */
	char *name;			/* Name of the set. */

	/*
	 * Pointer to chain of sub sets (eg, RAID10) which
	 * make up device-mapper stacks.
	 *
	 * Lowest level identified with NULL here.
	 */
	struct list_head sets;

	/*
	 * Pointer to the disks of the RAID volume.
	 *
	 * High sets in the stack will have NULL here.
	 *
	 * Lowest level will hold devices definitions for IDE/SCSI/... devices.
	 */
	struct list_head devs;

	unsigned int stride;		/* stripesize */
	enum type type;			/* Unified raid type */
	unsigned int has_partitions;	/* Flag to indicate RAID set	
					   has partitions */
	enum status status;		/* Status of set */

};

const char *get_type(struct lib_context *lc, enum type type);
const char *get_set_type(struct lib_context *lc, void *rs);
const char *get_status(struct lib_context *lc, void *rs);
uint64_t calc_total_sectors(struct lib_context *lc, struct dmraid_set *rs);
struct dmraid_dev *alloc_dev(struct lib_context *lc);
void free_dev(struct lib_context *lc, struct dmraid_dev *rd);
void list_add_dev_sorted(struct lib_context *lc, struct dmraid_set *rs,
			 struct dmraid_dev *new,
			 int (*f)(struct dmraid_dev *rd,
			 struct dmraid_dev *new));
struct dmraid_dev *list_del_dev(struct lib_context *lc, struct dmraid_set *rs);
struct dmraid_set *alloc_set(struct lib_context *lc);
unsigned int count_devs(struct lib_context *lc, struct dmraid_set *rs);
void free_set(struct lib_context *lc, struct dmraid_set *rs);
struct dmraid_set *find_set(struct lib_context *lc,
			    const char *name, int decend);
const char *get_set_name(struct lib_context *lc, void *rs);
int group_set(struct lib_context *lc, char *name);
void process_sets(int (*func)(struct lib_context *l, void *v, int a),
		  struct lib_context *lc, int arg, int partitions);
int write_set(struct lib_context *lc, void *rs);
int partitioned_set(struct lib_context *lc, void *rs);
int base_partitioned_set(struct lib_context *lc, void *rs);
void discover_raid_devices(struct lib_context *lc, char **fmt_names, char **devices);
void discover_partitions(struct lib_context *lc);
unsigned int count_devices(struct lib_context *lc, enum dev_type type);
enum type rd_type(struct types *types, unsigned int type);
void file_metadata(struct lib_context *lc, const char *handler,
		   char *name, void *data, size_t size, uint64_t offset);
int erase_metadata(struct lib_context *lc);

#endif
