/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Tool library
 */

#include <dmraid.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "commands.h"
#include "toollib.h"

/* [De]activate a RAID set. */
static int _activate_set(struct lib_context *lc, void *rs, int arg)
{
	/* FIXME: startup array daemon on activate; stop on deactivate */
	if ((ACTIVATE & action) ? activate_set(lc, rs) :
				  deactivate_set(lc, rs)) {
		log_notice(lc, "%sctivated %s RAID set %s",
			   action & ACTIVATE ? "A": "Dea", get_set_type(lc, rs),
			   get_set_name(lc, rs));
		return 1;
	}

	return 0;
}

/* [De]activate RAID sets. */
int activate_sets(struct lib_context *lc, int arg)
{
	/* Discover partitions to deactivate RAID sets for and work on them. */
	if (DEACTIVATE & action) {
		discover_partitions(lc);
		process_sets(_activate_set, lc, 0, 1);
	}

	process_sets(_activate_set, lc, arg, 0);

	/* Discover partitions to activate RAID sets for and work on them. */
	if ((ACTIVATE & action) && !(NO_PARTITIONS & action)) {
		discover_partitions(lc);
		process_sets(_activate_set, lc, 0, 1);
	}

	return 1;
}

/* Build all sets or the ones given. */
void build_sets(struct lib_context *lc, char **sets)
{
	int o = 0;

	do {
		if (!group_set(lc, sets[o]))
			log_err(lc, "failed building set");

		if (!sets[o])
			break;
	} while (sets[++o]);
}

/* Convert a character sting to lower case. */
void str_tolower(char *s)
{
	for (; *s; s++)
		*s = tolower(*s);
}

/* Check if selected or all formats shall be used to read the metadata. */
static int push_format(char ***fmt_names, char *name)
{
	unsigned int i;
	char **p = *fmt_names;

	for (i = 0; p && p[i]; )
		i++;

	if (!(p = dbg_realloc(p, (i + 2) * sizeof (*p))))
		return 0;

	p[i] = name;
	p[i+1] = NULL;
	*fmt_names = p;

	return 1;
}

void free_format_names(struct lib_context *lc)
{
	if (format_names) {
		unsigned int i = 0;

		for (i = 0; format_names[i]; i++)
			dbg_free(format_names[i]);

		dbg_free(format_names);
	}
}

int valid_format(struct lib_context *lc, char *fmt,
		 char ***fmt_names, char **invalid)
{
	int ret = 1;
	char *c, *name, *p = fmt;

	while (*p && ret) {
		if ((c = strchr(p, ',')))
			*c = '\0';

		if (!(name = dbg_strdup(p)))
			return 0;

		if (!(ret = check_valid_format(lc, name)))
			*invalid = name;

		if (!push_format(fmt_names, name))
			return 0;

		p += strlen(p);
		if (c) {
			*c = ',';
			p++;
		}
	}

	return ret;
}

void format_error(struct lib_context *lc, char **fmt_names, const char *error)
{
	if (fmt_names) {
		unsigned int i;

		log_print_nnl(lc, "No RAID %s with format: ", error);
		for (i = 0; fmt_names[i]; i++) {
			log_print_nnl(lc, "%s", fmt_names[i]);
			if (fmt_names[i+1])
				log_print_nnl(lc, ",");
		}
		log_print_nnl(lc, "\n");
	} else
		log_print(lc, "No Software RAID %ss", error);
}
